from __future__ import annotations
from typing import List
import panel as pn

from bencher.results.bench_result_base import EmptyContainer
from bencher.results.video_summary import VideoSummaryResult
from bencher.results.panel_result import PanelResult
from bencher.results.plotly_result import PlotlyResult
from bencher.results.holoview_result import HoloviewResult
from bencher.utils import listify


class BenchResult(PlotlyResult, HoloviewResult, VideoSummaryResult):
    """Contains the results of the benchmark and has methods to cast the results to various datatypes and graphical representations"""

    def __init__(self, bench_cfg) -> None:
        PlotlyResult.__init__(self, bench_cfg)
        HoloviewResult.__init__(self, bench_cfg)

    @staticmethod
    def default_plot_callbacks():
        return [
            # VideoSummaryResult.to_video_summary, #quite expensive so not turned on by default
            HoloviewResult.to_bar,
            HoloviewResult.to_scatter_jitter,
            HoloviewResult.to_curve,
            HoloviewResult.to_line,
            HoloviewResult.to_heatmap,
            PlotlyResult.to_volume,
            # PanelResult.to_video,
            PanelResult.to_panes,
        ]

    @staticmethod
    def plotly_callbacks():
        return [HoloviewResult.to_surface, PlotlyResult.to_volume]

    def plot(self) -> pn.panel:
        """Plots the benchresult using the plot callbacks defined by the bench run

        Returns:
             pn.panel: A panel representation of the results
        """
        if self.bench_cfg.plot_callbacks is not None:
            return pn.Column(*[cb(self) for cb in self.bench_cfg.plot_callbacks])
        return None

    def to_auto(
        self,
        plot_list: List[callable] = None,
        remove_plots: List[callable] = None,
        **kwargs,
    ) -> List[pn.panel]:
        self.plt_cnt_cfg.print_debug = False
        plot_list = listify(plot_list)
        remove_plots = listify(remove_plots)

        if plot_list is None:
            plot_list = BenchResult.default_plot_callbacks()
        if remove_plots is not None:
            for p in remove_plots:
                plot_list.remove(p)

        kwargs = self.set_plot_size(**kwargs)

        row = EmptyContainer(pn.Row())
        for plot_callback in plot_list:
            if self.plt_cnt_cfg.print_debug:
                print(f"checking: {plot_callback.__name__}")
            # the callbacks are passed from the static class definition, so self needs to be passed before the plotting callback can be called
            row.append(plot_callback(self, **kwargs))

        self.plt_cnt_cfg.print_debug = True
        if len(row.pane) == 0:
            row.append(
                pn.pane.Markdown("No Plotters are able to represent these results", **kwargs)
            )
        return row.pane

    def to_auto_plots(self, **kwargs) -> pn.panel:
        """Given the dataset result of a benchmark run, automatically dedeuce how to plot the data based on the types of variables that were sampled

        Args:
            bench_cfg (BenchCfg): Information on how the benchmark was sampled and the resulting data

        Returns:
            pn.pane: A panel containing plot results
        """

        plot_cols = pn.Column()
        plot_cols.append(self.to_sweep_summary(name="Plots View"))
        plot_cols.append(self.to_auto(**kwargs))
        plot_cols.append(self.bench_cfg.to_post_description())
        return plot_cols



from collections import namedtuple
import xarray as xr
from sortedcontainers import SortedDict
import hashlib
import re
import math
from colorsys import hsv_to_rgb
from pathlib import Path
from uuid import uuid4
from functools import partial
from typing import Callable, Any, List
import param


def hmap_canonical_input(dic: dict) -> tuple:
    """From a dictionary of kwargs, return a hashable representation (tuple) that is always the same for the same inputs and retains the order of the input arguments.  e.g, {x=1,y=2} -> (1,2) and {y=2,x=1} -> (1,2).  This is used so that keywords arguments can be hashed and converted the the tuple keys that are used for holomaps

    Args:
        dic (dict): dictionary with keyword arguments and values in any order

    Returns:
        tuple: values of the dictionary always in the same order and hashable
    """

    function_input = SortedDict(dic)
    return tuple(function_input.values())


def make_namedtuple(class_name: str, **fields) -> namedtuple:
    """Convenience method for making a named tuple

    Args:
        class_name (str): name of the named tuple

    Returns:
        namedtuple: a named tuple with the fields as values
    """
    return namedtuple(class_name, fields)(*fields.values())


def get_nearest_coords(dataset: xr.Dataset, collapse_list=False, **kwargs) -> dict:
    """Given an xarray dataset and kwargs of key value pairs of coordinate values, return a dictionary of the nearest coordinate name value pair that was found in the dataset

    Args:
        ds (xr.Dataset): dataset

    Returns:
        dict: nearest coordinate name value pair that matches the input coordinate name value pairs.
    """

    selection = dataset.sel(method="nearest", **kwargs)
    cd = selection.coords.to_dataset().to_dict()["coords"]
    cd2 = {}
    for k, v in cd.items():
        cd2[k] = v["data"]
        if collapse_list and isinstance(cd2[k], list):
            cd2[k] = cd2[k][0]  # select the first item in the list
    return cd2


def get_nearest_coords1D(val: Any, coords) -> Any:
    if isinstance(val, (int, float)):
        return min(coords, key=lambda x_: abs(x_ - val))
    return val


def hash_sha1(var: any) -> str:
    """A hash function that avoids the PYTHONHASHSEED 'feature' which returns a different hash value each time the program is run"""
    return hashlib.sha1(str(var).encode("ASCII")).hexdigest()


def capitalise_words(message: str):
    """Given a string of lowercase words, capitalise them

    Args:
        message (str): lower case string

    Returns:
        _type_: capitalised string
    """
    capitalized_message = " ".join([word.capitalize() for word in message.split(" ")])
    return capitalized_message


def un_camel(camel: str) -> str:
    """Given a snake_case string return a CamelCase string

    Args:
        camel (str): camelcase string

    Returns:
        str: uncamelcased string
    """

    return capitalise_words(re.sub("([a-z])([A-Z])", r"\g<1> \g<2>", camel.replace("_", " ")))


def int_to_col(int_val, sat=0.5, val=0.95, alpha=-1) -> tuple[float, float, float]:
    """Uses the golden angle to generate colors programmatically with minimum overlap between colors.
    https://martin.ankerl.com/2009/12/09/how-to-create-random-colors-programmatically/

    Args:
        int_val (_type_): index of an object you want to color, this is mapped to hue in HSV
        sat (float, optional): saturation in HSV. Defaults to 0.5.
        val (float, optional): value in HSV. Defaults to 0.95.
        alpha (int, optional): transparency.  If -1 then only RGB is returned, if 0 or greater, RGBA is returned. Defaults to -1.

    Returns:
        tuple[float, float, float] | tuple[float, float, float, float]: either RGB or RGBA vector
    """
    golden_ratio_conjugate = (1 + math.sqrt(5)) / 2
    rgb = hsv_to_rgb(int_val * golden_ratio_conjugate, sat, val)
    if alpha >= 0:
        return (*rgb, alpha)
    return rgb


def lerp(value, input_low: float, input_high: float, output_low: float, output_high: float):
    input_low = float(input_low)
    return output_low + ((float(value) - input_low) / (float(input_high) - input_low)) * (
        float(output_high) - output_low
    )


def color_tuple_to_css(color: tuple[float, float, float]) -> str:
    return f"rgb{(color[0] * 255, color[1] * 255, color[2] * 255)}"


def gen_path(filename, folder, suffix):
    path = Path(f"cachedir/{folder}/{filename}/")
    path.mkdir(parents=True, exist_ok=True)
    return f"{path.absolute().as_posix()}/{filename}_{uuid4()}{suffix}"


def gen_video_path(video_name: str = "vid", extension: str = ".webm") -> str:
    return gen_path(video_name, "vid", extension)


def gen_image_path(image_name: str = "img", filetype=".png") -> str:
    return gen_path(image_name, "img", filetype)


def callable_name(any_callable: Callable[..., Any]) -> str:
    if isinstance(any_callable, partial):
        return any_callable.func.__name__
    try:
        return any_callable.__name__
    except AttributeError:
        return str(any_callable)


def listify(obj) -> list:
    """Take an object and turn it into a list if its not already a list.  However if the object is none, don't turn it into a list"""
    if obj is None:
        return None
    if isinstance(obj, list):
        return obj
    if isinstance(obj, tuple):
        return list(obj)
    return [obj]


def get_name(var):
    if isinstance(var, param.Parameter):
        return var.name
    return var


def params_to_str(param_list: List[param.Parameter]):
    return [get_name(i) for i in param_list]



from enum import auto

from strenum import StrEnum

import bencher as bch


class ExampleEnum(StrEnum):
    value_1 = auto()
    value_2 = auto()
    # value3 = auto()
    # value4 = auto()


class Cfg(bch.ParametrizedSweep):
    enum1 = bch.EnumSweep(ExampleEnum)
    result = bch.ResultVar()

    # def __call__(self,**kwargs) -> Any:
    #     self.update_params_from_kwargs(**kwargs)
    #     self.result = float(str(self.enum1)[-1])
    #     return self.get_results_values_as_dict()


def bench_function(cfg: Cfg):
    return {"result": float(str(cfg.enum1)[-1])}


def print_assert_equal(msg, first, second):
    print(f"{msg} {first}=={second}")
    assert first == second


def assert_call_counts(bencher, run_cfg, wrapper_calls=-1, fn_calls=-1, cache_calls=-1):
    print_assert_equal(
        "worker wrapper call count",
        bencher.sample_cache.worker_wrapper_call_count,
        wrapper_calls * run_cfg.repeats,
    )
    print_assert_equal(
        "worker fn call count",
        bencher.sample_cache.worker_fn_call_count,
        fn_calls * run_cfg.repeats,
    )
    print_assert_equal(
        "worker cache call count",
        bencher.sample_cache.worker_cache_call_count,
        cache_calls * run_cfg.repeats,
    )


def example_cache_context() -> bch.Bench:
    run_cfg = bch.BenchRunCfg()
    run_cfg.use_sample_cache = True
    run_cfg.only_hash_tag = True
    run_cfg.repeats = 2
    run_cfg.parallel = False

    bencher = bch.Bench("bench_context", bench_function, Cfg, run_cfg=run_cfg)

    # clear all tags from the cache at the beginning so that the example works the same not matter how many times the example is run.  When using this for you own code you probably don't want to clear the cache at the beginning because you will lose all the data you collected.
    bencher.clear_tag_from_sample_cache("example_tag1", run_cfg)
    bencher.clear_tag_from_sample_cache("example_tag2", run_cfg)

    # run a benchmark with a constant value and save results with example_tag1
    bencher.plot_sweep(
        title="Benchmark enum=value_1",
        const_vars=[Cfg.param.enum1.with_const(ExampleEnum.value_1)],
        result_vars=[Cfg.param.result],
        tag="example_tag1",
    )

    # there are not values in the cache, so we expect 1 fn call and 0 cache calls
    assert_call_counts(bencher, run_cfg, wrapper_calls=1, fn_calls=1, cache_calls=0)

    # now run another benchmark with the same tag but a different value
    bencher.clear_call_counts()
    bencher.plot_sweep(
        title="Benchmark enum=value_2",
        const_vars=[Cfg.param.enum1.with_const(ExampleEnum.value_2)],
        result_vars=[Cfg.param.result],
        tag="example_tag1",
    )

    # these values have not been calcuated before so there should be 1 fn call
    assert_call_counts(bencher, run_cfg, wrapper_calls=1, fn_calls=1, cache_calls=0)

    # now create a new benchmark that calculates the values of the previous two benchmarks. The tag is the same so those values will be loaded from the cache instead of getting calculated again
    bencher.clear_call_counts()
    bencher.plot_sweep(
        title="Benchmark enum=[value_1,value_2] combined",
        input_vars=[Cfg.param.enum1],
        result_vars=[Cfg.param.result],
        tag="example_tag1",
    )

    # both calls hit the cache.
    assert_call_counts(bencher, run_cfg, wrapper_calls=2, fn_calls=0, cache_calls=2)

    # run the same benchmark as before but use a different tag.  The previously cached values will not be used and fresh values will be calculated instead
    bencher.clear_call_counts()
    bencher.plot_sweep(
        title="Benchmark enum=[value_1,value_2] with different tag",
        input_vars=[Cfg.param.enum1],
        result_vars=[Cfg.param.result],
        tag="example_tag2",
    )

    # Both calls are calcuated becuase the tag is different so they don't hit the cache
    assert_call_counts(bencher, run_cfg, wrapper_calls=2, fn_calls=2, cache_calls=0)

    return bencher


if __name__ == "__main__":
    example_cache_context().report.show()



from enum import Enum, auto
from typing import Any
from bencher.results.float_formatter import FormatFloat


# TODO enable these options
class ComposeType(Enum):
    right = auto()  # append the container to the right (creates a row)
    down = auto()  # append the container below (creates a column)
    overlay = auto()  # overlay on top of the current container (alpha blending)
    sequence = auto()  # display the container after (in time)


class ComposableContainerBase:
    """A base class for renderer backends.  A composable renderr"""

    @staticmethod
    def label_formatter(var_name: str, var_value: int | float | str) -> str:
        """Take a variable name and values and return a pretty version with approximate fixed width

        Args:
            var_name (str): The name of the variable, usually a dimension
            var_value (int | float | str): The value of the dimension

        Returns:
            str: Pretty string representation with fixed width
        """

        if isinstance(var_value, (int, float)):
            var_value = FormatFloat()(var_value)
        if var_name is not None and var_value is not None:
            return f"{var_name}={var_value}"
        if var_name is not None:
            return f"{var_name}"
        if var_value is not None:
            return f"{var_value}"
        return None

    def __init__(
        self, horizontal: bool = True, compose_method: ComposeType = ComposeType.right
    ) -> None:
        self.horizontal: bool = horizontal
        self.compose_method = compose_method
        self.container = []

    def append(self, obj: Any) -> None:
        """Add an object to the container.  The relationship between the objects is defined by the ComposeType

        Args:
            obj (Any): Object to add to the container
        """
        self.container.append(obj)

    def render(self):
        """Return a representation of the container that can be composed with other render() results. This function can also be used to defer layout and rending options until all the information about the container content is known.  You may need to ovverride this method depending on the container. See composable_container_video as an example.

        Returns:
            Any: Visual representation of the container that can be combined with other containers
        """
        return self.container



import bencher as bch
import numpy as np
import matplotlib.pyplot as plt


# code from https://ipython-books.github.io/124-simulating-a-partial-differential-equation-reaction-diffusion-systems-and-turing-patterns/
class TuringPattern(bch.ParametrizedSweep):
    alpha = bch.FloatSweep(default=2.8e-4, bounds=(2e-4, 5e-3))
    beta = bch.FloatSweep(default=5e-3, bounds=(1e-3, 9e-3))
    tau = bch.FloatSweep(default=0.1, bounds=(0.01, 0.5))
    k = bch.FloatSweep(default=-0.005, bounds=(-0.01, 0.01))

    size = bch.IntSweep(default=30, bounds=(30, 200), doc="size of the 2D grid")
    time = bch.FloatSweep(default=20.0, bounds=(1, 100), doc="total time of simulation")
    dt = bch.FloatSweep(default=0.001, doc="simulation time step")

    video = bch.ResultVideo()
    score = bch.ResultVar()

    def laplacian(self, Z, dx):
        Ztop = Z[0:-2, 1:-1]
        Zleft = Z[1:-1, 0:-2]
        Zbottom = Z[2:, 1:-1]
        Zright = Z[1:-1, 2:]
        Zcenter = Z[1:-1, 1:-1]
        return (Ztop + Zleft + Zbottom + Zright - 4 * Zcenter) / dx**2

    def update(self, U, V, dx):
        # We compute the Laplacian of u and v.
        deltaU = self.laplacian(U, dx)
        deltaV = self.laplacian(V, dx)
        # We take the values of u and v inside the grid.
        Uc = U[1:-1, 1:-1]
        Vc = V[1:-1, 1:-1]
        # We update the variables.
        U[1:-1, 1:-1], V[1:-1, 1:-1] = (
            Uc + self.dt * (self.alpha * deltaU + Uc - Uc**3 - Vc + self.k),
            Vc + self.dt * (self.beta * deltaV + Uc - Vc) / self.tau,
        )
        # Neumann conditions: derivatives at the edges
        # are null.
        for Z in (U, V):
            Z[0, :] = Z[1, :]
            Z[-1, :] = Z[-2, :]
            Z[:, 0] = Z[:, 1]
            Z[:, -1] = Z[:, -2]

    def __call__(self, **kwargs):
        self.update_params_from_kwargs(**kwargs)

        n = int(self.time / self.dt)  # number of iterations
        dx = 2.0 / self.size  # space step

        U = np.random.rand(self.size, self.size)
        V = np.random.rand(self.size, self.size)

        fig, ax = plt.subplots(frameon=False, figsize=(2, 2))
        fig.set_tight_layout(True)
        ax.set_axis_off()
        vid_writer = bch.VideoWriter()
        for i in range(n):
            self.update(U, V, dx)
            if i % 500 == 0:
                ax.imshow(U)
                fig.canvas.draw()
                rgb = np.array(fig.canvas.renderer.buffer_rgba())
                vid_writer.append(rgb)

        self.video = vid_writer.write()

        self.score = self.alpha + self.beta
        return super().__call__()


def example_video(
    run_cfg: bch.BenchRunCfg = bch.BenchRunCfg(), report: bch.BenchReport = bch.BenchReport()
) -> bch.Bench:
    # run_cfg.auto_plot = False
    # run_cfg.use_sample_cache = True
    bench = bch.Bench("example_video", TuringPattern(), run_cfg=run_cfg, report=report)

    bench.plot_sweep(
        "Turing patterns with different parameters",
        input_vars=[TuringPattern.param.alpha, TuringPattern.param.beta],
        # input_vars=[TuringPattern.param.alpha],
        result_vars=[TuringPattern.param.video],
    )

    return bench


def example_video_tap(
    run_cfg: bch.BenchRunCfg = bch.BenchRunCfg(), report: bch.BenchReport = bch.BenchReport()
) -> bch.Bench:  # pragma: no cover
    bench = TuringPattern().to_bench(run_cfg=run_cfg, report=report)
    bench.plot_sweep(
        input_vars=["alpha", "beta"],
        # result_vars=["video","score"],
    )

    return bench


if __name__ == "__main__":
    run_cfg_ex = bch.BenchRunCfg()
    run_cfg_ex.level = 2
    run_cfg_ex.use_sample_cache = True

    # example_video(run_cfg_ex).report.show()
    example_video_tap(run_cfg_ex).report.show()



from datetime import datetime
from typing import List

from pandas import Timestamp
from param import Selector
from bencher.variables.sweep_base import SweepBase, shared_slots


class TimeBase(SweepBase, Selector):
    """A class to capture a time snapshot of benchmark values.  Time is reprented as a continous value i.e a datetime which is converted into a np.datetime64.  To represent time as a discrete value use the TimeEvent class. The distinction is because holoview and plotly code makes different assumptions about discrete vs continous variables"""

    __slots__ = shared_slots

    def values(self) -> List[str]:
        """return all the values for a parameter sweep.  If debug is true return a reduced list"""
        # print(self.sampling_str(debug))
        return self.objects


class TimeSnapshot(TimeBase):
    """A class to capture a time snapshot of benchmark values.  Time is reprented as a continous value i.e a datetime which is converted into a np.datetime64.  To represent time as a discrete value use the TimeEvent class. The distinction is because holoview and plotly code makes different assumptions about discrete vs continous variables"""

    __slots__ = shared_slots

    def __init__(
        self,
        datetime_src: datetime | str,
        units: str = "time",
        samples: int = None,
        **params,
    ):
        if isinstance(datetime_src, str):
            TimeBase.__init__(self, [datetime_src], instantiate=True, **params)
        else:
            TimeBase.__init__(
                self,
                objects=[Timestamp(datetime_src)],
                instantiate=True,
                **params,
            )
        self.units = units
        if samples is None:
            self.samples = len(self.objects)
        else:
            self.samples = samples


class TimeEvent(TimeBase):
    """A class to represent a discrete event in time where the data was captured i.e a series of pull requests.  Here time is discrete and can't be interpolated, to represent time as a continous value use the TimeSnapshot class.  The distinction is because holoview and plotly code makes different assumptions about discrete vs continous variables"""

    __slots__ = shared_slots

    def __init__(
        self,
        time_event: str,
        units: str = "event",
        samples: int = None,
        **params,
    ):
        TimeBase.__init__(
            self,
            objects=[time_event],
            instantiate=True,
            **params,
        )
        self.units = units
        if samples is None:
            self.samples = len(self.objects)
        else:
            self.samples = samples



from functools import partial
from typing import List, Tuple, Any
from param import Parameter, Parameterized
import holoviews as hv
import panel as pn


from bencher.utils import make_namedtuple, hash_sha1
from bencher.variables.results import (
    ResultVar,
    ResultVec,
    ResultHmap,
    ResultVideo,
    ResultImage,
    ResultString,
    ResultContainer,
    ResultReference,
)


class ParametrizedSweep(Parameterized):
    """Parent class for all Sweep types that need a custom hash"""

    @staticmethod
    def param_hash(param_type: Parameterized, hash_value: bool = True) -> int:
        """A custom hash function for parametrised types with options for hashing the value of the type and hashing metadata

        Args:
            param_type (Parameterized): A parameter
            hash_value (bool, optional): use the value as part of the hash. Defaults to True.
            # hash_meta (bool, optional): use metadata as part of the hash. Defaults to False.

        Returns:
            int: a hash
        """

        curhash = 0
        if hash_value:
            for k, v in param_type.param.values().items():
                if k != "name":
                    curhash = hash_sha1((curhash, hash_sha1(v)))

        # if hash_meta:
        #     for k, v in param_type.param.objects().items():
        #         if k != "name":
        #             print(f"key:{k}, hash:{hash_sha1(k)}")
        #             print(f"value:{v}, hash:{hash_sha1(v)}")
        #             curhash = hash_sha1((curhash, hash_sha1(k), hash_sha1(v)))
        return curhash

    def hash_persistent(self) -> str:
        """A hash function that avoids the PYTHONHASHSEED 'feature' which returns a different hash value each time the program is run"""
        return ParametrizedSweep.param_hash(self, True)

    def update_params_from_kwargs(self, **kwargs) -> None:
        """Given a dictionary of kwargs, set the parameters of the passed class 'self' to the values in the dictionary."""
        used_params = {}
        for key in self.param.objects().keys():
            if key in kwargs:
                if key != "name":
                    used_params[key] = kwargs[key]

        self.param.update(**used_params)

    @classmethod
    def get_input_and_results(cls, include_name: bool = False) -> Tuple[dict, dict]:
        """Get dictionaries of input parameters and result parameters

        Args:
            cls: A parametrised class
            include_name (bool): Include the name parameter that all parametrised classes have. Default False

        Returns:
            Tuple[dict, dict]: a tuple containing the inputs and result parameters as dictionaries
        """
        inputs = {}
        results = {}
        for k, v in cls.param.objects().items():
            if isinstance(
                v,
                (
                    ResultVar,
                    ResultVec,
                    ResultHmap,
                    ResultVideo,
                    ResultImage,
                    ResultString,
                    ResultContainer,
                    ResultReference,
                ),
            ):
                results[k] = v
            else:
                inputs[k] = v

        if not include_name:
            inputs.pop("name")
        return make_namedtuple("inputresult", inputs=inputs, results=results)

    def get_inputs_as_dict(self) -> dict:
        """Get the key:value pairs for all the input variables"""
        inp = self.get_input_and_results().inputs
        vals = self.param.values()
        return {i: vals[i] for i, v in inp.items()}

    def get_results_values_as_dict(self, holomap=None) -> dict:
        """Get a dictionary of result variables with the name and the current value"""
        values = self.param.values()
        output = {key: values[key] for key in self.get_input_and_results().results}
        if holomap is not None:
            output |= {"hmap": holomap}
        return output

    @classmethod
    def get_inputs_only(cls) -> List[Parameter]:
        """Return a list of input parameters

        Returns:
            List[param.Parameter]: A list of input parameters
        """
        return list(cls.get_input_and_results().inputs.values())

    @staticmethod
    def filter_fn(item, p_name):
        return item.name != p_name

    @classmethod
    def get_input_defaults(
        cls,
        override_defaults: List = None,
    ) -> List[Tuple[Parameter, Any]]:
        inp = cls.get_inputs_only()
        if override_defaults is None:
            override_defaults = []
        assert isinstance(override_defaults, list)

        for p in override_defaults:
            inp = filter(partial(ParametrizedSweep.filter_fn, p_name=p[0].name), inp)

        return override_defaults + [[i, i.default] for i in inp]

    @classmethod
    def get_input_defaults_override(cls, **kwargs) -> dict[str, Any]:
        inp = cls.get_inputs_only()
        defaults = {}
        for i in inp:
            defaults[i.name] = i.default

        for k, v in kwargs.items():
            defaults[k] = v

        return defaults

    @classmethod
    def get_results_only(cls) -> List[Parameter]:
        """Return a list of result parameters

        Returns:
            List[param.Parameter]: A list of result parameters
        """
        return list(cls.get_input_and_results().results.values())

    @classmethod
    def get_inputs_as_dims(
        self, compute_values=False, remove_dims: str | List[str] = None
    ) -> List[hv.Dimension]:
        inputs = self.get_inputs_only()

        if remove_dims is not None:
            if isinstance(remove_dims, str):
                remove_dims = [remove_dims]
            filtered_inputs = [i for i in inputs if i.name not in remove_dims]
            inputs = filtered_inputs

        return [iv.as_dim(compute_values) for iv in inputs]

    def to_dynamic_map(
        self, callback=None, name=None, remove_dims: str | List[str] = None, result_var: str = None
    ) -> hv.DynamicMap:
        if callback is None:
            callback = self.__call__

        if result_var is None:
            result_vars = self.get_input_and_results().results
            for k, rv in result_vars.items():
                if isinstance(rv, ResultHmap):
                    result_var = k

        def callback_wrapper(**kwargs):
            return callback(**kwargs)[result_var]

        return hv.DynamicMap(
            callback=callback_wrapper,
            kdims=self.get_inputs_as_dims(compute_values=False, remove_dims=remove_dims),
            name=name,
        ).opts(shared_axes=False, framewise=True, width=1000, height=1000)

    def to_gui(self, result_var: str = None, **kwargs):  # pragma: no cover
        main = pn.Row(
            self.to_dynamic_map(result_var=result_var, **kwargs),
        )
        main.show()

    def to_holomap(self, callback, remove_dims: str | List[str] = None) -> hv.DynamicMap:
        return hv.HoloMap(
            hv.DynamicMap(
                callback=callback,
                kdims=self.get_inputs_as_dims(compute_values=True, remove_dims=remove_dims),
            )
        )

    def __call__(self):
        return self.get_results_values_as_dict()

    def plot_hmap(self, **kwargs):
        return self.__call__(**kwargs)["hmap"]

    def to_bench(self, run_cfg=None, report=None, name: str = None):
        from bencher import Bench

        assert isinstance(self, ParametrizedSweep)

        if name is None:
            name = self.name[:-5]  # param adds 5 digit number to the end, so remove it

        return Bench(name, self, run_cfg=run_cfg, report=report)



from bencher.utils import params_to_str
from typing import Optional, List
import itertools
import panel as pn
import xarray as xr
from param import Parameter
from bencher.results.bench_result_base import BenchResultBase, ReduceType
from bencher.variables.results import ResultImage
from bencher.plotting.plot_filter import VarRange, PlotFilter
from bencher.utils import callable_name, listify
from bencher.video_writer import VideoWriter
from bencher.results.float_formatter import FormatFloat
from bencher.results.video_result import VideoControls
from bencher.utils import int_to_col
from bencher.results.composable_container.composable_container_video import ComposableContainerVideo


class VideoSummaryResult(BenchResultBase):
    def to_video_summary(
        self,
        result_var: Parameter = None,
        input_order: List[str] = None,
        reverse: bool = True,
        result_types=(ResultImage,),
        **kwargs,
    ) -> Optional[pn.panel]:
        plot_filter = PlotFilter(
            float_range=VarRange(0, None),
            cat_range=VarRange(0, None),
            panel_range=VarRange(1, None),
        )
        matches_res = plot_filter.matches_result(
            self.plt_cnt_cfg, callable_name(self.to_video_summary_ds)
        )

        # video_controls = VideoControls()
        if matches_res.overall:
            ds = self.to_dataset(ReduceType.SQUEEZE)
            row = pn.Row()
            for rv in self.get_results_var_list(result_var):
                if isinstance(rv, result_types):
                    row.append(self.to_video_summary_ds(ds, rv, input_order, reverse, **kwargs))
            return row
        return matches_res.to_panel()

    def to_video_summary_ds(
        self,
        dataset: xr.Dataset,
        result_var: Parameter,
        input_order: List[str] = None,
        reverse: bool = True,
        video_controls: VideoControls = None,
        **kwargs,
    ):
        vr = VideoWriter()
        da = dataset[result_var.name]

        if input_order is None:
            input_order = list(da.dims)
        else:
            input_order = params_to_str(input_order)
        if reverse:
            input_order = list(reversed(input_order))

        inputs_produc = [da.coords[i].values for i in input_order]

        for index in itertools.product(*inputs_produc):
            lookup = dict(zip(input_order, index))
            val = da.loc[lookup].item()
            index = listify(index)
            for i in range(len(index)):
                if isinstance(index[i], (int, float)):
                    index[i] = FormatFloat()(index[i])
            label = ", ".join(f"{a[0]}={a[1]}" for a in list(zip(input_order, index)))
            if val is not None:
                vr.append_file(val, label)
        fn = vr.write_png(**kwargs)
        kwargs.pop("target_duration", None)
        if fn is not None:
            if video_controls is None:
                video_controls = VideoControls()
            vid = video_controls.video_container(fn, **kwargs)
            return vid

        return None

    def to_video_grid(
        self,
        result_var: Parameter = None,
        result_types=(ResultImage,),
        pane_collection: pn.pane = None,
        **kwargs,
    ) -> Optional[pn.panel]:
        """Returns the results compiled into a video

        Args:
            result_var (Parameter, optional): The result var to plot. Defaults to None.
            result_types (tuple, optional): The types of result var to convert to video. Defaults to (ResultImage,).
            collection (pn.pane, optional): If there are multiple results, use this collection to stack them. Defaults to pn.Row().

        Returns:
            Optional[pn.panel]: a panel pane with a video of all results concatenated together
        """
        plot_filter = PlotFilter(
            float_range=VarRange(0, None),
            cat_range=VarRange(0, None),
            panel_range=VarRange(1, None),
            input_range=VarRange(1, None),
        )
        matches_res = plot_filter.matches_result(
            self.plt_cnt_cfg, callable_name(self.to_video_grid_ds)
        )

        if pane_collection is None:
            pane_collection = pn.Row()

        if matches_res.overall:
            ds = self.to_dataset(ReduceType.SQUEEZE)
            for rv in self.get_results_var_list(result_var):
                if isinstance(rv, result_types):
                    pane_collection.append(self.to_video_grid_ds(ds, rv, **kwargs))
            return pane_collection
        return matches_res.to_panel()

    def to_video_grid_ds(
        self,
        dataset: xr.Dataset,
        result_var: Parameter,
        reverse=True,
        video_controls: VideoControls = None,
        target_duration: float = None,
        **kwargs,
    ):
        vr = VideoWriter()

        cvc = self._to_video_panes_ds(
            dataset,
            self.plot_cb,
            target_dimension=0,
            horizontal=True,
            result_var=result_var,
            final=True,
            reverse=reverse,
            target_duration=target_duration,
            **kwargs,
        )

        fn = vr.write_video_raw(cvc)

        if fn is not None:
            if video_controls is None:
                video_controls = VideoControls()
            vid = video_controls.video_container(fn, **kwargs)
            return vid
        return None

    def plot_cb(self, dataset, result_var, **kwargs):
        val = self.ds_to_container(dataset, result_var, container=None, **kwargs)
        return val

    def _to_video_panes_ds(
        self,
        dataset: xr.Dataset,
        plot_callback: callable = None,
        target_dimension=0,
        horizontal=False,
        result_var=None,
        final=False,
        reverse=False,
        target_duration: float = None,
        **kwargs,
    ) -> pn.panel:
        num_dims = len(dataset.sizes)
        dims = list(d for d in dataset.sizes)
        if reverse:
            dims = list(reversed(dims))

        if num_dims > (target_dimension) and num_dims != 0:
            selected_dim = dims[-1]
            # print(f"selected dim {selected_dim}")
            dim_color = int_to_col(num_dims - 2, 0.05, 1.0)
            # sliced = dataset.isel({selected_dim: i})
            # label_val = sliced.coords[selected_dim].values.item()

            outer_container = ComposableContainerVideo(
                name=" vs ".join(dims),
                background_col=dim_color,
                horizontal=horizontal,
                target_duration=target_duration,
                # var_name=selected_dim,
                # var_value=label_val,
            )
            max_len = 0
            for i in range(dataset.sizes[selected_dim]):
                sliced = dataset.isel({selected_dim: i})
                label_val = sliced.coords[selected_dim].values.item()
                inner_container = ComposableContainerVideo(
                    outer_container.name,
                    var_name=selected_dim,
                    var_value=label_val,
                    horizontal=horizontal,
                    target_duration=target_duration,
                )
                panes = self._to_video_panes_ds(
                    sliced,
                    plot_callback=plot_callback,
                    target_dimension=target_dimension,
                    horizontal=len(sliced.sizes) <= target_dimension + 1,
                    result_var=result_var,
                )
                inner_container.append(panes)

                max_len = max(max_len, inner_container.label_len)
                rendered = inner_container.render()
                outer_container.append(rendered)
            return outer_container.render(concatenate=final)
        return plot_callback(dataset=dataset, result_var=result_var, **kwargs)



from typing import Protocol, Callable, List
import logging
from bencher.bench_cfg import BenchRunCfg, BenchCfg
from bencher.variables.parametrised_sweep import ParametrizedSweep
from bencher.bencher import Bench
from bencher.bench_report import BenchReport
from copy import deepcopy


class Benchable(Protocol):
    def bench(self, run_cfg: BenchRunCfg, report: BenchReport) -> BenchCfg:
        raise NotImplementedError


class BenchRunner:
    """A class to manage running multiple benchmarks in groups, or running the same benchmark but at multiple resolutions"""

    def __init__(
        self,
        name: str,
        bench_class=None,
        run_cfg: BenchRunCfg = BenchRunCfg(),
        publisher: Callable = None,
    ) -> None:
        self.name = name
        self.run_cfg = BenchRunner.setup_run_cfg(run_cfg)
        self.bench_fns = []
        self.publisher = publisher
        if bench_class is not None:
            self.add_bench(bench_class)
        self.results = []
        self.servers = []

    @staticmethod
    def setup_run_cfg(
        run_cfg: BenchRunCfg = BenchRunCfg(), level: int = 2, use_cache=True
    ) -> BenchRunCfg:
        run_cfg_out = deepcopy(run_cfg)
        run_cfg_out.use_sample_cache = use_cache
        run_cfg_out.only_hash_tag = use_cache
        run_cfg_out.level = level
        return run_cfg_out

    @staticmethod
    def from_parametrized_sweep(
        class_instance: ParametrizedSweep,
        run_cfg: BenchRunCfg = BenchRunCfg(),
        report: BenchReport = BenchReport(),
    ):
        return Bench(f"bench_{class_instance.name}", class_instance, run_cfg=run_cfg, report=report)

    def add_run(self, bench_fn: Benchable) -> None:
        self.bench_fns.append(bench_fn)

    def add_bench(self, class_instance: ParametrizedSweep) -> None:
        def cb(run_cfg: BenchRunCfg, report: BenchReport) -> BenchCfg:
            bench = BenchRunner.from_parametrized_sweep(
                class_instance, run_cfg=run_cfg, report=report
            )
            return bench.plot_sweep(f"bench_{class_instance.name}")

        self.add_run(cb)

    def run(
        self,
        min_level: int = 2,
        max_level: int = 6,
        level: int = None,
        repeats: int = 1,
        run_cfg: BenchRunCfg = None,
        publish: bool = False,
        debug: bool = False,
        show: bool = False,
        save: bool = False,
        grouped: bool = True,
        use_cache: bool = True,
    ) -> List[Bench]:
        """This function controls how a benchmark or a set of benchmarks are run. If you are only running a single benchmark it can be simpler to just run it directly, but if you are running several benchmarks together and want them to be sampled at different levels of fidelity or published together in a single report this function enables that workflow.  If you have an expensive function, it can be useful to view low fidelity results as they are computed but also continue to compute higher fidelity results while reusing previously computed values. The parameters min_level and max_level let you specify how to progressivly increase the sampling resolution of the benchmark sweep. By default use_cache=True so that previous values are reused.

        Args:
            min_level (int, optional): The minimum level to start sampling at. Defaults to 2.
            max_level (int, optional): The maximum level to sample up to. Defaults to 6.
            level (int, optional): If this is set, then min_level and max_level are not used and only a single level is sampled. Defaults to None.
            repeats (int, optional): The number of times to run the entire benchmarking procedure. Defaults to 1.
            run_cfg (BenchRunCfg, optional): benchmark run configuration. Defaults to None.
            publish (bool, optional): Publish the results to git, requires a publish url to be set up. Defaults to False.
            debug (bool, optional): _description_. Defaults to False.
            show (bool, optional): show the results in the local web browswer. Defaults to False.
            save (bool, optional): save the results to disk in index.html. Defaults to False.
            grouped (bool, optional): Produce a single html page with all the benchmarks included. Defaults to True.
            use_cache (bool, optional): Use the sample cache to reused previous results. Defaults to True.

        Returns:
            List[BenchCfg]: A list of bencher instances
        """
        if run_cfg is None:
            run_cfg = deepcopy(self.run_cfg)
        run_cfg = BenchRunner.setup_run_cfg(run_cfg, use_cache=use_cache)

        if level is not None:
            min_level = level
            max_level = level
        for r in range(1, repeats + 1):
            for lvl in range(min_level, max_level + 1):
                if grouped:
                    report_level = BenchReport(f"{run_cfg.run_tag}_{self.name}")

                for bch_fn in self.bench_fns:
                    run_lvl = deepcopy(run_cfg)
                    run_lvl.level = lvl
                    run_lvl.repeats = r
                    logging.info(f"Running {bch_fn} at level: {lvl} with repeats:{r}")
                    if grouped:
                        res = bch_fn(run_lvl, report_level)
                    else:
                        res = bch_fn(run_lvl, BenchReport())
                        res.report.bench_name = f"{run_cfg.run_tag}_{res.report.bench_name}"
                        self.show_publish(res.report, show, publish, save, debug)
                    self.results.append(res)
                if grouped:
                    self.show_publish(report_level, show, publish, save, debug)
        return self.results

    def show_publish(self, report, show, publish, save, debug):
        if save:
            report.save_index()
        if publish and self.publisher is not None:
            report.publish(remote_callback=self.publisher, debug=debug)
        if show:
            self.servers.append(report.show())

    def shutdown(self):
        while self.servers:
            self.servers.pop().stop()

    def __del__(self) -> None:
        self.shutdown()



# you need this import to be able to reference a class from a static method in that class
from __future__ import annotations

import math
import random
import time
from datetime import datetime
from enum import auto

from strenum import StrEnum

import bencher as bch


# define a class with the output variables you want to benchmark. It must inherit from ParametrizedSweep (which inherits from param.Parametrized). Param is a python library that allows you to track metadata about parameters.  I would recommend reading at least the intro: https://param.holoviz.org/.  I have extended param with some extra metadata such is the units of the variable so that it can automaticaly be plotted.
class OutputCfg(bch.ParametrizedSweep):
    """A class for defining what variables the benchmark function returns and metadata on those variables"""

    # Documenting the variable here with enables automatic summaries of what has been benchmarked.
    # This made up example uses accuracy as an example, but the variable defined here can be any metric that is important for the performance of a system.  You can also define the direction of the optimisation i.e. to minimise or maximise the metric.
    accuracy = bch.ResultVar(
        units="%", direction=bch.OptDir.maximize, doc="The accuracy of the algorithm."
    )


# Define categorical variables with enums that inherit from StrEnum.  In this example, its just an arbitrary set of categories that have an unknown influence on the metric you want to understand. In a real world case these would be a set of conditions or settings you are benchmarking
class AlgoSetting(StrEnum):
    """Use enums to describe categorical input.  In this example they are given names that describe how they affect the function output, but in a real world example these will be some settings to an algorithm that you want to understand how they affect the metric you are trying to optimise."""

    # add some random noise to the output.  When your algorithm has noisy output it often is an indication that something is not quite right.  The graphs should show that you want to avoid the "noisy" setting in your algorithm
    noisy = auto()

    # This is the setting with the best performance, and characterising that that is is the goal of the benchmarking
    optimum = auto()

    poor = auto()  # this setting results in poor performance


# define a class with the input variables you want to benchmark.  It must inherit from ParametrizeSweep. This class defines a struct that is passed to the benchmark function.  The function must be pure and so we define it as a staticmethod that takes an InputCfg class and returns an OutputCfg class. By accepting and returning parametrized classes the metadata about what the relationship between the input and output are easy to track.
class InputCfg(bch.ParametrizedSweep):
    # The variables must be defined as one of the Sweep types, i.e, FloatSweep, IntSweep, EnumSweep from bencher.bench_vars
    # theta = FloatSweep(default=0, bounds=[0, math.pi], doc="Input angle", units="rad", samples=30)

    # Define sweep variables by passing in an enum class name. The first element of the enum is the default by convention, but you can overrride the default in the constructor
    algo_setting_enum = bch.EnumSweep(AlgoSetting, default=AlgoSetting.poor)

    # In this case there are no units so its marked as unitless or ul. You can define how many evenly distributed samples to sample the parameter with
    algo_setting_float = bch.FloatSweep(
        default=0.0,
        bounds=[0.0, 6.0],
        doc="This represents a continuous input value to your function that affects the desired output in a way you want to characterise.",
        units="ul",
        samples=10,
    )

    # define the objective function you want to benchmark. It must be static and have no side effects. It should accept 1 input of type InputCfg (or whatever your input config class is called) and return the OutputCfg class you have defined
    @staticmethod
    def bench_function(cfg: InputCfg) -> OutputCfg:
        """Takes an ExampleBenchCfgIn and returns a ExampleBenchCfgOut output.  This is just a dummy example so the behavior of the function is rather transparent, but in a real use case the function would be a black box you want to characterise."""
        output = OutputCfg()

        output.accuracy = 50 + math.sin(cfg.algo_setting_float) * 5

        # this simulates random long term change in the function
        output.accuracy += time.localtime(datetime.now().second).tm_sec / 30

        match cfg.algo_setting_enum:
            case AlgoSetting.noisy:
                # add some random noise to the output.  When your algorith has noisy output it often is an indication that something is not quite right.  The graphs should show that you want to avoid the "noisy" setting in your algorithm
                output.accuracy += random.uniform(-10, 10)
            case AlgoSetting.optimum:
                output.accuracy += 30  # This is the setting with the best performance, and characterising that is is the goal of the benchmarking
            case AlgoSetting.poor:
                output.accuracy -= 20  # this setting results in poor performance
        return output


if __name__ == "__main__":
    # pass the objective function you have defined to bencher.  This benchmark function can be reused for multiple sweeps.  You also need to pass the inputCfg to the bencher so that it can process the metadata about the input configuration.
    bench = bch.Bench("Bencher_Example_Categorical", InputCfg.bench_function, InputCfg)

    # Bencher needs to know the metadata of the variable in order to automatically sweep and plot it, so it is passed by using param's metadata syntax.  InputCfg.param.* is how to access the metadata defined in the class description.  Unfortunately vscode autocomplete doesn't work with params metaclass machinery so you will need to look at the class definition to get a list of possible settings. Define what parameter you want to sweep over and the result variable you want to plot.  If you pass 1 input, it will perform a 1D sweep over that dimension and plot a line or a bar graph of the result (depending on if that variable on continuous or discrete).  In this example we are going to sweep the enum variable and record the accuracy.
    bench.plot_sweep(
        input_vars=[InputCfg.param.algo_setting_enum],
        result_vars=[OutputCfg.param.accuracy],
        title="Simple example 1D enum sweep",
        description="""Sample all the values in enum setting and record the resulting accuracy.  The algo_setting_float is not mentioned in the inputs and so it takes the default value that was set in the InputCfg class.  Repeats=10 so the benchmark function is called 10 times serially.  This is why the function must be pure, if a past call to the function affects the future call to the function (through global side effects) any statistics you calculate will not be correct. 
        """,
        post_description="Here you can see the affect of each setting on the output and the optimum is clearly the best.",
        run_cfg=bch.BenchRunCfg(repeats=10),
    )

    # There is also a floating point input setting that affects the performance of the algorithm.  By passing only the float setting, the InputCfg class will use the default setting of the categorical value so you can understand the float setting in isolation
    bench.plot_sweep(
        input_vars=[InputCfg.param.algo_setting_float],
        result_vars=[OutputCfg.param.accuracy],
        title="Simple example 1D float sweep",
        description="""Perform a 1D sweep over the continuous variable algo_setting_float taking sweep the bounds and number of samples from the InputCfg class definition.  The algo_setting_enum is not mentioned in the inputs and so it takes the default value that was set in the InputCfg class.  Repeats=10 so the benchmark function is called 10 times serially.  
        """,
        post_description="The plot shows the output is affected by the float input in a continuous way with a peak around 1.5",
        run_cfg=bch.BenchRunCfg(repeats=10),
    )

    # This sweep is a combination of the previous two sweeps
    bench.plot_sweep(
        input_vars=[
            InputCfg.param.algo_setting_float,
            InputCfg.param.algo_setting_enum,
        ],
        result_vars=[OutputCfg.param.accuracy],
        title="Simple example 2D sweep",
        description="""Perform a 2D sweep over the enum and continuous variable to see how they act together.  Here the setting use_optuna=True so additional graphs a plotted at the end. 
        """,
        post_description="In this example function the two input settings combine in a linear and predictable way, so the best combination of settings is enum = AlgoSetting.optimum and float = 1.33.  Setting use_optuna=True adds a plot of how much each input parameter affects the metric and a printout of the best parameter values found during the sweep.  If the value for repeat is high it is an indication there is something wrong with your benchmark function. The repeat should have no affect on the value of the function if calls to the function are independent.  This can be useful to detect undesired side effects in your code",
        run_cfg=bch.BenchRunCfg(repeats=10, use_optuna=True, serve_xarray=True, serve_pandas=True),
    )

    # In the last example we track the value of the categorical values over time.
    # run this code in a loop twice to simulate calling the benchmarking function at different times.  The most common use case for tracking over time would be run once a day during nightly benchmarking
    bench.plot_sweep(
        input_vars=[InputCfg.param.algo_setting_enum],
        result_vars=[OutputCfg.param.accuracy],
        const_vars=[(InputCfg.param.algo_setting_float, 1.33)],
        title="Simple example 1D sweep over time",
        description="""Once you have found the optimal settings for your algorithm you want to make sure that the performance is not lost over time.  You can set variables to a constant value and in this case the float value is set to its optimum value.  The first time this function is run only the results from sweeping the categorical value is plotted (the same as example 1), but the second time it is run a graph the values over time is shown. [Run the code again if you don't see a graph over time]. If the graphs over time shows long term changes (not just noise), it indicate there is another external factor that is affecting your performace over time, i.e. dependencies changing, physical degradation of equipment, an unnoticed bug from a pull request etc...

        This shows the basic features of bencher.  These examples are purposefully simplified to demonstrate its features in isolation and don't reeally show the real advantages of bencher.  If you only have a few inputs and outputs its not that complicated to throw together some plots of performance.  The power of bencher is that when you have a system with many moving parts that all interact with eachother, teasing apart those influences becomes much harder because the parameter spaces combine quite quickly into a high dimensional mess. Bencher makes it easier to experiment with different combination of inputs to gain an intuition of the system performance. Bencher can plot up to 6D input natively and you can add custom plots if you have exotic data types or state spaces [WIP]. 
        """,
        post_description="",
        run_cfg=bch.BenchRunCfg(repeats=10, over_time=True, clear_history=False),
    )

    # launch web server and view
    bench.report.show()



# pylint: disable=duplicate-code


import bencher as bch

# All the examples will be using the data structures and benchmark function defined in this file
from bencher.example.benchmark_data import ExampleBenchCfgIn, ExampleBenchCfgOut, ExampleBenchCfg


def example_floats(
    run_cfg: bch.BenchRunCfg = bch.BenchRunCfg(), report: bch.BenchReport = bch.BenchReport()
) -> bch.Bench:
    """Example of how to perform a parameter sweep for floating point variables

    Args:
        run_cfg (BenchRunCfg): configuration of how to perform the param sweep

    Returns:
        Bench: results of the parameter sweep
    """
    bench = bch.Bench("Bencher_Example_Floats", ExampleBenchCfg(), report=report, run_cfg=run_cfg)

    with open("README.md", "r", encoding="utf-8") as file:
        readme = file.read()

    bench.report.append(readme, "Intro")

    bench.plot_sweep(
        input_vars=[ExampleBenchCfgIn.param.theta],
        result_vars=[ExampleBenchCfgOut.param.out_sin],
        title="Float 1D Example",
        description="""Bencher is a tool to make it easy to explore how input parameter affect a range of output metrics.  In these examples we are going to benchmark an example function which has been selected to show the features of bencher.
        The example function takes an input theta and returns the absolute value of sin(theta) and cos(theta) +- various types of noise.

        def bench_function(cfg: ExampleBenchCfgIn) -> ExampleBenchCfgOut:
            "Takes an ExampleBenchCfgIn and returns a ExampleBenchCfgOut output"
            out = ExampleBenchCfgOut()
            noise = calculate_noise(cfg)
            offset = 0.0

            postprocess_fn = abs if cfg.postprocess_fn == PostprocessFn.absolute else negate_fn

            out.out_sin = postprocess_fn(offset + math.sin(cfg.theta) + noise)
            out.out_cos = postprocess_fn(offset + math.cos(cfg.theta) + noise)
            return out

    The following examples will show how to perform parameter sweeps to characterise the behavior of the function.  The idea is that the benchmarking can be used to gain understanding of an unknown function. 
        """,
        post_description="Here you can see the output plot of sin theta between 0 and pi.  In the tabs at the top you can also view 3 tabular representations of the data",
    )

    bench.plot_sweep(
        input_vars=[ExampleBenchCfgIn.param.theta, ExampleBenchCfgIn.param.noisy],
        result_vars=[ExampleBenchCfgOut.param.out_sin],
        title="Float 1D and Bool Example",
        description="""Following from the previous example lets add another input parameter to see how that affects the output.  We pass the boolean  'noisy' and keep the other parameters the same""",
        post_description="Now the plot has two lines, one for each of the boolean values where noisy=true and noisy=false.",
    )

    bench.plot_sweep(
        input_vars=[ExampleBenchCfgIn.param.theta, ExampleBenchCfgIn.param.noisy],
        result_vars=[ExampleBenchCfgOut.param.out_sin, ExampleBenchCfgOut.param.out_cos],
        title="Float 1D and Bool Example with multiple outputs",
        description="""Following from the previous example here the second output is added to the result variables""",
        post_description="Another column is added for the result variable that shows cos(theta)",
    )

    bench.plot_sweep(
        input_vars=[
            ExampleBenchCfgIn.param.theta,
            ExampleBenchCfgIn.param.noisy,
            ExampleBenchCfgIn.param.postprocess_fn,
        ],
        result_vars=[
            ExampleBenchCfgOut.param.out_sin,
            ExampleBenchCfgOut.param.out_cos,
        ],
        title="Float 1D, Bool and Categorical Example",
        description="""Following from the previous example lets add another input parameter to see how that affects the output.  We add the 'postprocess_fn' categorical enum value which either takes the absolute value or negates the output of the function.""",
        post_description="This generates two rows of results, one for each of the category options.",
    )

    return bench


if __name__ == "__main__":
    bench_ex = example_floats(bch.BenchRunCfg(repeats=2))
    bench_ex.report.save_index()
    bench_ex.report.show()



# import math

# import bencher as bch


# def bench_function(
#     theta: float = 0,
#     offset: float = 0,
#     scale: float = 1.0,
#     trig_func: str = "sin",
#     **kwargs,  # pylint: disable=unused-argument
# ) -> dict:
#     """All the other examples use classes and parameters to define the inputs and outputs to the function. However it makes the code less flexible when integrating with other systems, so this example shows a more basic interface that accepts and returns dictionaries.  The classes still need to be defined however because that is how the sweep and plotting settings are calcuated"""
#     output = {}

#     if trig_func == "sin":
#         output["voltage"] = offset + math.sin(theta) * scale
#     elif trig_func == "cos":
#         output["voltage"] = offset + math.cos(theta) * scale

#     return output


# class InputCfg(bch.ParametrizedSweep):
#     """This class is used to define the default values and bounds of the variables to benchmark."""

#     theta = bch.FloatSweep(
#         default=0.0,
#         bounds=[0.0, 6.0],
#         doc="Input angle to the trig function",
#         units="rad",
#         samples=10,
#     )

#     offset = bch.FloatSweep(
#         default=0.0,
#         bounds=[0.0, 3.0],
#         doc="Add an offset voltage to the result of the trig function",
#         units="v",
#         samples=5,
#     )

#     trig_func = bch.StringSweep(["sin", "cos"], doc="Select what trigonometric function use")


# class OutputVoltage(bch.ParametrizedSweep):
#     voltage = bch.ResultVar(units="v", doc="Output voltage")


# if __name__ == "__main__":
#     # pass the objective function you have defined to bencher.  The other examples pass the InputCfg type, but this benchmark function accepts a kwargs dictionary so you don't need to pass the inputCfg type.
#     bench = bch.Bench("Bencher_Example_Categorical", bench_function)

#     # Bencher needs to know the metadata of the variable in order to automatically sweep and plot it, so it is passed by using param's metadata syntax.  InputCfg.param.* is how to access the metadata defined in the class description.
#     bench.plot_sweep(
#         input_vars=[InputCfg.param.theta, InputCfg.param.offset],
#         result_vars=[OutputVoltage.param.voltage],
#         title="Example with kwarg inputs and dict output",
#         description=bench_function.__doc__,
#     )

#     # launch web server and view
#     bench.report.show()



"""This file contains an example of how to define benchmarking parameters sweeps. Categorical values are defined as enums and passed to EnumSweep classes, other types of sweeps are defined by their respective classes.

You can define a subclass which contains an input configuration which can be passed to a function in a type safe way. You can combine the subclass with a higher level class which contains more configuation parameters.  This is to help manage the complexity of large configuration/parameter spaces. 
"""

import math
import random
from enum import auto

from strenum import StrEnum


from bencher.variables.inputs import IntSweep, FloatSweep, StringSweep, EnumSweep, BoolSweep
from bencher.variables.results import ResultVar, OptDir

from bencher.variables.parametrised_sweep import ParametrizedSweep


class PostprocessFn(StrEnum):
    """Apply a postprocessing step to the data"""

    absolute = auto()  # return the abs of the output data
    negate = auto()  # return the negative of the output data


class NoiseDistribution(StrEnum):
    """A categorical variable describing the types of random noise"""

    uniform = auto()  # uniform random noiase
    gaussian = auto()  # gaussian noise
    lognorm = auto()  # lognorm noise


class NoiseCfg(ParametrizedSweep):
    """A class for storing the parameters for generating various types of noise"""

    noisy = BoolSweep(
        default=False, doc="Optionally add random noise to the output of the function"
    )

    noise_distribution = EnumSweep(NoiseDistribution, doc=NoiseDistribution.__doc__)

    sigma = FloatSweep(
        default=1,
        bounds=[0, 10],
        doc="The standard deviation of the noise",
        units="v",
    )


def calculate_noise(config: NoiseCfg) -> float:
    """Generate a float value based on a noise distribution and scale

    Args:
        config (NoiseCfg): see NoiseCfg type

    Returns:
        float: a noisy float value
    """

    noise = 0.0
    if config.noisy:
        match config.noise_distribution:
            case NoiseDistribution.uniform:
                noise = random.uniform(0, config.sigma)
            case NoiseDistribution.gaussian:
                noise = random.gauss(0, config.sigma)
            case NoiseDistribution.lognorm:
                noise = random.lognormvariate(0, config.sigma)

    return noise


class ExampleBenchCfgIn(NoiseCfg):
    """A class for representing a set of configuration options for the example worker. This class inherits from NoiseCfg so it also stores all its values as well."""

    theta = FloatSweep(default=0, bounds=[0, math.pi], doc="Input angle", units="rad", samples=30)
    offset = FloatSweep(default=0, bounds=[0, 0.3], doc="dc offset", units="v", samples=30)
    postprocess_fn = EnumSweep(PostprocessFn)


class ExampleBenchCfgOut(ParametrizedSweep):
    out_sin = ResultVar(units="v", direction=OptDir.minimize, doc="sin of theta with some noise")
    out_cos = ResultVar(units="v", direction=OptDir.minimize, doc="cos of theta with some noise")
    out_bool = ResultVar(units="%", doc="sin > 0.5")


def negate_fn(fn_input: float):
    """returns the negative of the input

    Args:
        fn_input (float): any float value

    Returns:
        float: negative of the input
    """
    return -fn_input


def bench_function(cfg: ExampleBenchCfgIn) -> ExampleBenchCfgOut:
    """Takes an ExampleBenchCfgIn and returns a ExampleBenchCfgOut output"""
    out = ExampleBenchCfgOut()
    noise = calculate_noise(cfg)

    postprocess_fn = abs if cfg.postprocess_fn == PostprocessFn.absolute else negate_fn

    out.out_sin = postprocess_fn(cfg.offset + math.sin(cfg.theta) + noise)
    out.out_cos = postprocess_fn(cfg.offset + math.cos(cfg.theta) + noise)

    out.out_bool = out.out_sin > 0.5
    return out


class ExampleBenchCfg(ParametrizedSweep):
    theta = FloatSweep(default=0, bounds=[0, math.pi], doc="Input angle", units="rad", samples=30)
    offset = FloatSweep(default=0, bounds=[0, 0.3], doc="dc offset", units="v", samples=30)
    postprocess_fn = EnumSweep(PostprocessFn)

    noisy = BoolSweep(
        default=False, doc="Optionally add random noise to the output of the function"
    )
    noise_distribution = EnumSweep(NoiseDistribution, doc=NoiseDistribution.__doc__)
    sigma = FloatSweep(
        default=1,
        bounds=[0, 10],
        doc="The standard deviation of the noise",
        units="v",
    )

    out_sin = ResultVar(units="v", direction=OptDir.minimize, doc="sin of theta with some noise")
    out_cos = ResultVar(units="v", direction=OptDir.minimize, doc="cos of theta with some noise")
    out_bool = ResultVar(units="%", doc="sin > 0.5")

    def __call__(self, **kwwargs) -> dict:
        self.update_params_from_kwargs(**kwwargs)

        noise = self.calculate_noise()
        postprocess_fn = abs if self.postprocess_fn == PostprocessFn.absolute else negate_fn

        self.out_sin = postprocess_fn(self.offset + math.sin(self.theta) + noise)
        self.out_cos = postprocess_fn(self.offset + math.cos(self.theta) + noise)
        self.out_bool = self.out_sin > 0.5
        return self.get_results_values_as_dict()

    def calculate_noise(self):
        noise = 0.0
        if self.noisy:
            match self.noise_distribution:
                case NoiseDistribution.uniform:
                    noise = random.uniform(0, self.sigma)
                case NoiseDistribution.gaussian:
                    noise = random.gauss(0, self.sigma)
                case NoiseDistribution.lognorm:
                    noise = random.lognormvariate(0, self.sigma)

        return noise


def call(**kwargs) -> dict:
    return ExampleBenchCfg().__call__(**kwargs)


class AllSweepVars(ParametrizedSweep):
    """A class containing all the sweep types, This class is used for unit testing how the configuration classes are serialised and hashed"""

    var_float = FloatSweep(default=5, bounds=(0, 10), units="m/s")
    var_int = IntSweep(default=3, bounds=[0, 4])
    var_int_big = IntSweep(default=0, bounds=[0, 100], samples=3)
    var_bool = BoolSweep()
    var_string = StringSweep(["string1", "string2"])
    var_enum = EnumSweep(PostprocessFn)

    result = ResultVar()

    def __call__(self, **kwargs) -> dict:
        self.update_params_from_kwargs(**kwargs)
        self.result = self.var_float + self.var_int
        return self.get_results_values_as_dict()


class SimpleBenchClass(ParametrizedSweep):
    var1 = IntSweep(default=0, bounds=[0, 2])

    result = ResultVar()

    def __call__(self, **kwargs) -> dict:
        self.update_params_from_kwargs(**kwargs)
        self.result = self.var1
        return self.get_results_values_as_dict()


class SimpleBenchClassFloat(ParametrizedSweep):
    var1 = FloatSweep(bounds=[0, 100])

    result = ResultVar()

    def __call__(self, **kwargs) -> dict:
        self.update_params_from_kwargs(**kwargs)
        self.result = self.var1
        return self.get_results_values_as_dict()



"""This file has some examples for how to perform basic benchmarking parameter sweeps"""

import bencher as bch

# All the examples will be using the data structures and benchmark function defined in this file
from bencher.example.benchmark_data import ExampleBenchCfgIn, ExampleBenchCfgOut, bench_function


def example_1D_float(
    run_cfg: bch.BenchRunCfg = bch.BenchRunCfg(), report: bch.BenchReport = bch.BenchReport()
) -> bch.Bench:
    """This example shows how to sample a 1 dimensional float variable and plot the result of passing that parameter sweep to the benchmarking function"""

    bench = bch.Bench(
        "benchmarking_example_float1D",
        bench_function,
        ExampleBenchCfgIn,
        run_cfg=run_cfg,
        report=report,
    )

    # here we sample the input variable theta and plot the value of output1. The (noisy) function is sampled 20 times so you can see the distribution
    bench.plot_sweep(
        title="Example 1D Float",
        input_vars=[ExampleBenchCfgIn.param.theta],
        result_vars=[ExampleBenchCfgOut.param.out_sin],
        description=example_1D_float.__doc__,
    )

    return bench


if __name__ == "__main__":
    ex_run_cfg = bch.BenchRunCfg()
    ex_run_cfg.repeats = 2
    # ex_run_cfg.over_time = True

    example_1D_float(ex_run_cfg).report.show()



# import numpy as np

# import bencher as bch


# class VolumeSample(bch.ParametrizedSweep):
#     """A class to represent a 3D point in space."""

#     pos_samples = 1

#     x = bch.FloatSweep(
#         default=0, bounds=[-1.0, 1.0], samples=pos_samples, doc="x coordinate of the sample volume"
#     )
#     y = bch.FloatSweep(
#         default=0, bounds=[-1.0, 1.0], samples=pos_samples, doc="y coordinate of the sample volume"
#     )
#     z = bch.FloatSweep(
#         default=0, bounds=[-1.0, 1.0], samples=pos_samples, doc="z coordinate of the sample volume"
#     )

#     vec_samples = 5

#     vec_x = bch.FloatSweep(
#         default=0, bounds=[-1.0, 1.0], samples=vec_samples, doc="x coordinate of the sample volume"
#     )
#     vec_y = bch.FloatSweep(
#         default=0, bounds=[-1.0, 1.0], samples=vec_samples, doc="y coordinate of the sample volume"
#     )
#     vec_z = bch.FloatSweep(
#         default=0, bounds=[-1.0, 1.0], samples=vec_samples, doc="z coordinate of the sample volume"
#     )


# class VolumeResult(bch.ParametrizedSweep):
#     """A class to represent the properties of a volume sample."""

#     vec_dir = bch.ResultVec(3, "vec", doc="A vector field with an interesting shape")


# def bench_fn(point: VolumeSample, normalise=False) -> VolumeResult:
#     """This function takes a 3D point as input and returns distance of that point to the origin.

#     Args:
#         point (VolumeSample): Sample point

#     Returns:
#         VolumeResult: Value at that point
#     """
#     output = VolumeResult()

#     vec = np.array([point.vec_x, point.vec_y, point.vec_z])

#     if normalise:
#         norm = np.linalg.norm(vec)
#         if norm > 0:
#             vec /= norm

#     output.vec_dir = list(vec)
#     return output


# def example_cone(run_cfg: bch.BenchRunCfg) -> bch.Bench:
#     """Example of how to perform a 3D floating point parameter sweep

#     Args:
#         run_cfg (BenchRunCfg): configuration of how to perform the param sweep

#     Returns:
#         Bench: results of the parameter sweep
#     """
#     bench = bch.Bench("Bencher_Example_Cone", bench_fn, VolumeSample)

#     bench.plot_sweep(
#         input_vars=[
#             VolumeSample.param.x,
#             VolumeSample.param.y,
#             VolumeSample.param.z,
#             VolumeSample.param.vec_x,
#             VolumeSample.param.vec_y,
#             VolumeSample.param.vec_z,
#         ],
#         result_vars=[
#             VolumeResult.param.vec_dir,
#         ],
#         title="Float 3D cone Example",
#         description="""This example shows how to sample 3 floating point variables and plot a vector field representation of the results.  The benchmark function returns the distance to the origin""",
#         run_cfg=run_cfg,
#     )

#     return bench


# if __name__ == "__main__":
#     ex_run_cfg = bch.BenchRunCfg()
#     ex_run_cfg.use_cache = True
#     example_cone(ex_run_cfg).report.show()



# THIS IS NOT A WORKING EXAMPLE YET

# pylint: disable=duplicate-code,unused-argument


import bencher as bch
import math
import random
import numpy as np
import holoviews as hv

from strenum import StrEnum
from enum import auto


class Function(StrEnum):
    fn_cos = auto()
    fn_sin = auto()
    fn_log = auto()
    fn_arctan = auto()

    def call(self, arg) -> float:
        """Calls the function defined by the name of the enum

        Returns:
            float: The result of calling the function defined by the enum
        """
        return getattr(np, self.removeprefix("fn_"))(arg)


class PlotFunctions(bch.ParametrizedSweep):
    phase = bch.FloatSweep(
        default=0, bounds=[0, math.pi], doc="Input angle", units="rad", samples=5
    )

    freq = bch.FloatSweep(default=1, bounds=[0, math.pi], doc="Input angle", units="rad", samples=5)

    theta = bch.FloatSweep(
        default=0, bounds=[0, math.pi], doc="Input angle", units="rad", samples=10
    )

    compute_fn = bch.EnumSweep(Function)

    # RESULT VARS
    fn_output = bch.ResultVar(units="v", doc="sin of theta with some noise")
    out_sum = bch.ResultVar(units="v", doc="The sum")
    ref = bch.ResultReference()
    holomap = bch.ResultHmap()

    def __call__(self, plot=True, **kwargs) -> dict:
        self.update_params_from_kwargs(**kwargs)
        noise = 0.1

        self.fn_output = self.compute_fn.call(self.phase + self.freq * self.theta) + random.uniform(
            0, noise
        )

        self.holomap = self.plot_holo(plot)
        self.ref = bch.ResultReference(self.ref)

        return self.get_results_values_as_dict()

    def plot_holo(self, plot=True) -> hv.core.ViewableElement:
        """Plots a generic representation of the object that is not a basic hv datatype. In this case its an image of the values of the object, but it could be any representation of the object, e.g. a screenshot of the object state"""
        if plot:
            pt = hv.Text(0, 0, f"{self.phase}\n{self.freq}\n {self.theta}")
            pt *= hv.Ellipse(0, 0, 1)
            return pt
        return None


def example_holosweep(
    run_cfg: bch.BenchRunCfg = bch.BenchRunCfg(), report: bch.BenchReport = bch.BenchReport()
) -> bch.Bench:
    bench = PlotFunctions().to_bench(run_cfg, report)

    bench.plot_sweep(
        input_vars=["theta", "freq"],
        result_vars=["fn_output", "holomap"],
    )

    # print("best", res.get_best_trial_params(True))
    # print(res.hmap_kdims)
    # bench.report.append(res.describe_sweep())
    # bench.report.append(res.to_optuna_plots())
    # bench.report.append(res.get_best_holomap())
    # bench.report.append(res.to_curve(), "Slider view")
    # bench.report.append(res.to_holomap())

    # bench.report.append(res.to_holomap().layout())
    return bench


if __name__ == "__main__":
    PlotFunctions().to_gui()
    bench_run = bch.BenchRunner("bench_runner_test")
    bench_run.add_run(example_holosweep)
    bench_run.run(level=6, show=True, use_cache=False)



import panel as pn
import plotly.graph_objs as go
from typing import Optional
import xarray as xr

from param import Parameter

from bencher.plotting.plot_filter import VarRange
from bencher.results.bench_result_base import BenchResultBase, ReduceType
from bencher.variables.results import ResultVar


class PlotlyResult(BenchResultBase):
    def to_volume(self, result_var: Parameter = None, **kwargs):
        return self.filter(
            self.to_volume_da,
            float_range=VarRange(3, 3),
            cat_range=VarRange(-1, 0),
            reduce=ReduceType.REDUCE,
            target_dimension=3,
            result_var=result_var,
            result_types=(ResultVar),
            **kwargs,
        )

    def to_volume_da(
        self, dataset: xr.Dataset, result_var: Parameter, width=600, height=600
    ) -> Optional[pn.pane.Plotly]:
        """Given a benchCfg generate a 3D surface plot
        Returns:
            pn.pane.Plotly: A 3d volume plot as a holoview in a pane
        """
        x = self.bench_cfg.input_vars[0]
        y = self.bench_cfg.input_vars[1]
        z = self.bench_cfg.input_vars[2]
        opacity = 0.1
        meandf = dataset[result_var.name].to_dataframe().reset_index()
        data = [
            go.Volume(
                x=meandf[x.name],
                y=meandf[y.name],
                z=meandf[z.name],
                value=meandf[result_var.name],
                isomin=meandf[result_var.name].min(),
                isomax=meandf[result_var.name].max(),
                opacity=opacity,
                surface_count=20,
            )
        ]

        layout = go.Layout(
            title=f"{result_var.name} vs ({x.name} vs {y.name} vs {z.name})",
            width=width,
            height=height,
            margin=dict(t=50, b=50, r=50, l=50),
            scene=dict(
                xaxis_title=f"{x.name} [{x.units}]",
                yaxis_title=f"{y.name} [{y.units}]",
                zaxis_title=f"{z.name} [{z.units}]",
            ),
        )

        fig = dict(data=data, layout=layout)

        return pn.pane.Plotly(fig, name="volume_plotly")



import numpy as np
import moviepy.video.io.ImageSequenceClip
from pathlib import Path
from .utils import gen_video_path, gen_image_path

import moviepy
from moviepy.editor import (
    VideoFileClip,
    ImageClip,
    ImageSequenceClip,
    clips_array,
    concatenate_videoclips,
)
from PIL import Image, ImageDraw


class VideoWriter:
    def __init__(self, filename: str = "vid") -> None:
        self.images = []
        self.image_files = []
        self.video_files = []
        self.filename = gen_video_path(filename)

    def append(self, img):
        self.images.append(img)

    def write(self) -> str:
        clip = moviepy.video.io.ImageSequenceClip.ImageSequenceClip(
            self.images, fps=30, with_mask=False, load_images=True
        )
        self.write_video_raw(clip)
        return self.filename

    @staticmethod
    def create_label(label, width=None, height=14):
        if width is None:
            width = len(label) * 8
        new_img = Image.new("RGB", (width, height), (255, 255, 255))
        # ImageDraw.Draw(new_img).text((width/2, 0), label, (0, 0, 0),align="center",achor="ms")
        ImageDraw.Draw(new_img).text((width / 2.0, 0), label, (0, 0, 0), anchor="mt", font_size=12)

        return new_img

    @staticmethod
    def label_image(path: Path, label, padding=20) -> Path:
        image = Image.open(path)
        new_img = VideoWriter.create_label(label, image.size[0], image.size[1] + padding)
        new_img.paste(image, (0, padding))
        return new_img

    def append_file(self, filepath, label=None):
        if label is not None:
            path = Path(filepath)
            new_path = path.with_name(path.stem + "_labelled" + path.suffix).as_posix()
            padding = 20
            match path.suffix:
                case ".png" | ".jpg":
                    image = Image.open(filepath)
                    new_img = self.create_label(label, image.size[0], image.size[1] + padding)
                    new_img.paste(image, (0, padding))
                    new_img.save(new_path)
                    self.image_files.append(new_path)
                case ".webm":
                    import warnings

                    video_clip = VideoFileClip(filepath)
                    new_img = self.create_label(label, video_clip.w, padding)

                    # Convert PIL image to MoviePy clip
                    label_clip = ImageClip(np.array(new_img), duration=video_clip.duration)

                    labeled_video_clip = clips_array([[label_clip], [video_clip]])

                    # otherwise ffmpeg complains that the file is not getting read. We don't need the file just the size
                    with warnings.catch_warnings():
                        warnings.simplefilter(action="ignore")
                        labeled_video_clip.write_videofile(new_path, remove_temp=True, logger=None)
                    self.video_files.append(new_path)
        else:
            self.image_files.append(filepath)

    def to_images_sequence(self, images, target_duration: float = 10.0, frame_time=None, **kwargs):
        target_duration = kwargs.pop("target_duration", target_duration)
        if isinstance(images, list) and len(images) > 0:
            if frame_time is None:
                fps = len(images) / target_duration
                fps = max(fps, 1)  # never slower that 1 seconds per frame
                fps = min(fps, 30)
            else:
                fps = 1.0 / frame_time
            return ImageSequenceClip(images, fps=fps, with_mask=False)
        return None

    def write_png(self, **kwargs):
        clip = None
        if len(self.image_files) > 0:
            clip = self.to_images_sequence(self.image_files, **kwargs)
        if len(self.video_files) > 0:
            clip = concatenate_videoclips([VideoFileClip(f) for f in self.video_files])
        if clip is not None:
            clip.write_videofile(self.filename)
            return self.filename
        return None

    def write_video_raw(self, video_clip: moviepy.video.VideoClip, fps: int = 30) -> str:
        video_clip.write_videofile(
            self.filename,
            codec="libvpx",
            audio=False,
            bitrate="0",
            fps=fps,
            ffmpeg_params=["-crf", "34"],
        )
        video_clip.close()
        return self.filename


def add_image(np_array: np.ndarray, name: str = "img"):
    filename = gen_image_path(name)
    Image.fromarray(np_array).save(filename)
    return filename



from enum import Enum
from typing import List, Any

import numpy as np
from param import Integer, Number, Selector
from bencher.variables.sweep_base import SweepBase, shared_slots


class SweepSelector(Selector, SweepBase):
    """A class to reprsent a parameter sweep of bools"""

    __slots__ = shared_slots

    def __init__(self, units: str = "ul", samples: int = None, **params):
        SweepBase.__init__(self)
        Selector.__init__(self, **params)

        self.units = units
        if samples is None:
            self.samples = len(self.objects)
        else:
            self.samples = samples

    def values(self) -> List[Any]:
        """return all the values for a parameter sweep.  If debug is true return a reduced list"""
        return self.indices_to_samples(self.samples, self.objects)


class BoolSweep(SweepSelector):
    """A class to reprsent a parameter sweep of bools"""

    def __init__(self, units: str = "ul", samples: int = None, default=True, **params):
        SweepSelector.__init__(
            self,
            units=units,
            samples=samples,
            default=default,
            objects=[True, False] if default else [False, True],
            **params,
        )


class StringSweep(SweepSelector):
    """A class to reprsent a parameter sweep of strings"""

    def __init__(
        self,
        string_list: List[str],
        units: str = "",
        samples: int = None,
        **params,
    ):
        SweepSelector.__init__(
            self,
            objects=string_list,
            instantiate=True,
            units=units,
            samples=samples,
            **params,
        )


class EnumSweep(SweepSelector):
    """A class to reprsent a parameter sweep of enums"""

    __slots__ = shared_slots

    def __init__(self, enum_type: Enum | List[Enum], units=" ", samples=None, **params):
        # The enum can either be an Enum type or a list of enums
        list_of_enums = isinstance(enum_type, list)
        selector_list = enum_type if list_of_enums else list(enum_type)
        SweepSelector.__init__(
            self,
            objects=selector_list,
            instantiate=True,
            units=units,
            samples=samples,
            **params,
        )
        if not list_of_enums:  # Grab the docs from the enum type def
            self.doc = enum_type.__doc__


class IntSweep(Integer, SweepBase):
    """A class to reprsent a parameter sweep of ints"""

    __slots__ = shared_slots + ["sample_values"]

    def __init__(self, units="ul", samples=None, sample_values=None, **params):
        SweepBase.__init__(self)
        Integer.__init__(self, **params)

        self.units = units

        if sample_values is None:
            if samples is None:
                if self.bounds is None:
                    raise RuntimeError("You must define bounds for integer types")
                self.samples = 1 + self.bounds[1] - self.bounds[0]
            else:
                self.samples = samples
            self.sample_values = None
        else:
            self.sample_values = sample_values
            self.samples = len(self.sample_values)
            if "default" not in params:
                self.default = sample_values[0]

    def values(self) -> List[int]:
        """return all the values for a parameter sweep.  If debug is true return the  list"""
        sample_values = (
            self.sample_values
            if self.sample_values is not None
            else list(range(int(self.bounds[0]), int(self.bounds[1] + 1)))
        )

        return self.indices_to_samples(self.samples, sample_values)

    ###THESE ARE COPIES OF INTEGER VALIDATION BUT ALSO ALLOW NUMPY INT TYPES
    def _validate_value(self, val, allow_None):
        if callable(val):
            return

        if allow_None and val is None:
            return

        if not isinstance(val, (int, np.integer)):
            raise ValueError(
                "Integer parameter %r must be an integer, " "not type %r." % (self.name, type(val))
            )

    ###THESE ARE COPIES OF INTEGER VALIDATION BUT ALSO ALLOW NUMPY INT TYPES
    def _validate_step(self, val, step):
        if step is not None and not isinstance(step, (int, np.integer)):
            raise ValueError(
                "Step can only be None or an " "integer value, not type %r" % type(step)
            )


class FloatSweep(Number, SweepBase):
    """A class to represent a parameter sweep of floats"""

    __slots__ = shared_slots + ["sample_values"]

    def __init__(self, units="ul", samples=10, sample_values=None, step=None, **params):
        SweepBase.__init__(self)
        Number.__init__(self, step=step, **params)

        self.units = units

        self.sample_values = sample_values

        if sample_values is None:
            self.samples = samples
        else:
            self.samples = len(self.sample_values)
            if "default" not in params:
                self.default = sample_values[0]

    def values(self) -> List[float]:
        """return all the values for a parameter sweep.  If debug is true return a reduced list"""
        samps = self.samples
        if self.sample_values is None:
            if self.step is None:
                return np.linspace(self.bounds[0], self.bounds[1], samps)

            return np.arange(self.bounds[0], self.bounds[1], self.step)
        return self.sample_values


def box(name, center, width):
    var = FloatSweep(default=center, bounds=(center - width, center + width))
    var.name = name
    return var


def with_level(arr: list, level) -> list:
    return IntSweep(sample_values=arr).with_level(level).values()
    # return tmp.with_sample_values(arr).with_level(level).values()



import panel as pn
from bencher.results.composable_container.composable_container_base import ComposableContainerBase


class ComposableContainerPanel(ComposableContainerBase):
    def __init__(
        self,
        name: str = None,
        var_name: str = None,
        var_value: str = None,
        width: int = None,
        background_col: str = None,
        horizontal: bool = True,
    ) -> None:
        super().__init__(horizontal)

        container_args = {
            "name": name,
            "styles": {},
        }

        self.name = name

        if width is not None:
            container_args["styles"]["border-bottom"] = f"{width}px solid grey"
        if background_col is not None:
            container_args["styles"]["background"] = background_col

        if horizontal:
            self.container = pn.Column(**container_args)
            align = ("center", "center")
        else:
            self.container = pn.Row(**container_args)
            align = ("end", "center")

        label = self.label_formatter(var_name, var_value)
        if label is not None:
            self.label_len = len(label)
            side = pn.pane.Markdown(label, align=align)
            self.append(side)



from .bencher import Bench, BenchCfg, BenchRunCfg
from .bench_runner import BenchRunner
from .example.benchmark_data import ExampleBenchCfgIn, ExampleBenchCfgOut, bench_function
from .bench_plot_server import BenchPlotServer
from .variables.sweep_base import hash_sha1
from .variables.inputs import IntSweep, FloatSweep, StringSweep, EnumSweep, BoolSweep, SweepBase
from .variables.time import TimeSnapshot

from .variables.inputs import box
from .variables.results import (
    ResultVar,
    ResultVec,
    ResultHmap,
    ResultVideo,
    ResultImage,
    ResultString,
    ResultContainer,
    ResultReference,
    ResultVolume,
    OptDir,
    curve,
)

from .plotting.plot_filter import VarRange, PlotFilter
from .utils import (
    hmap_canonical_input,
    get_nearest_coords,
    make_namedtuple,
    gen_path,
    gen_image_path,
    gen_video_path,
    lerp,
)
from .variables.parametrised_sweep import ParametrizedSweep
from .caching import CachedParams
from .results.bench_result import BenchResult
from .results.panel_result import PanelResult
from .results.holoview_result import ReduceType, HoloviewResult
from .bench_report import BenchReport
from .job import Executors
from .video_writer import VideoWriter, add_image
from .class_enum import ClassEnum, ExampleEnum



import numpy as np
from moviepy.editor import (
    ImageClip,
    CompositeVideoClip,
    clips_array,
    concatenate_videoclips,
    VideoClip,
)

from bencher.results.composable_container.composable_container_base import ComposableContainerBase
from bencher.video_writer import VideoWriter


class ComposableContainerVideo(ComposableContainerBase):
    def __init__(
        self,
        name: str = None,
        var_name: str = None,
        var_value: str = None,
        background_col: tuple[3] = (255, 255, 255),
        horizontal: bool = True,
        target_duration: float = None,
    ) -> None:
        super().__init__(horizontal)
        self.name = name
        self.container = []
        self.background_col = background_col
        self.target_duration = 10 if target_duration is None else target_duration
        self.var_name = var_name

        self.label = self.label_formatter(var_name, var_value)
        if self.label is not None:
            self.label_len = len(self.label)

        # label = self.label_formatter(var_name, var_value)
        # if label is not None:
        # self.label_len = len(label)
        # side = pn.pane.Markdown(label, align=align)
        # self.append(side)

    def append(self, obj: VideoClip | str) -> None:
        if isinstance(obj, VideoClip):
            self.container.append(obj)
        else:
            # if self.label is not None:
            # img_obj = np.array(VideoWriter.label_image(obj, self.label))
            # else:
            # img_obj = obj
            self.container.append(ImageClip(obj))

    def render(self, concatenate: bool = False) -> CompositeVideoClip:
        fps = len(self.container) / self.target_duration
        fps = max(fps, 1.0)  # never slower that 1 seconds per frame
        fps = min(fps, 30.0)

        for i in range(len(self.container)):
            self.container[i].duration = 1.0 / fps
        if concatenate:
            out = concatenate_videoclips(self.container)
        else:
            if self.horizontal:
                clips = [self.container]
            else:
                clips = [[c] for c in self.container]
            out = clips_array(clips, bg_color=self.background_col)

        if self.label is not None:
            label = ImageClip(np.array(VideoWriter.create_label(self.label)))
            con2 = ComposableContainerVideo(
                background_col=self.background_col,
                horizontal=not self.horizontal,
            )
            con2.append(label)
            con2.append(out)
            return con2.render()
        return out



from __future__ import annotations
import logging
from typing import List, Optional
import panel as pn
import holoviews as hv
from param import Parameter
from functools import partial
import hvplot.xarray  # noqa pylint: disable=duplicate-code,unused-import
import xarray as xr

from bencher.utils import hmap_canonical_input, get_nearest_coords, get_nearest_coords1D
from bencher.results.panel_result import PanelResult
from bencher.results.bench_result_base import ReduceType

from bencher.plotting.plot_filter import PlotFilter, VarRange
from bencher.variables.results import ResultVar


hv.extension("bokeh", "plotly")


class HoloviewResult(PanelResult):
    @staticmethod
    def set_default_opts(width=600, height=600):
        width_heigh = {"width": width, "height": height, "tools": ["hover"]}
        hv.opts.defaults(
            hv.opts.Curve(**width_heigh),
            hv.opts.Points(**width_heigh),
            hv.opts.Bars(**width_heigh),
            hv.opts.Scatter(**width_heigh),
            hv.opts.HeatMap(cmap="plasma", **width_heigh, colorbar=True),
            # hv.opts.Surface(**width_heigh),
            hv.opts.GridSpace(plot_size=400),
        )
        return width_heigh

    def to(self, hv_type: hv.Chart, reduce: ReduceType = ReduceType.AUTO, **kwargs) -> hv.Chart:
        return self.to_hv_dataset(reduce).to(hv_type, **kwargs)

    def overlay_plots(self, plot_callback: callable) -> Optional[hv.Overlay]:
        results = []
        markdown_results = pn.Row()
        for rv in self.bench_cfg.result_vars:
            res = plot_callback(rv)
            if res is not None:
                if isinstance(res, pn.pane.Markdown):
                    markdown_results.append(res)
                else:
                    results.append(res)
        if len(results) > 0:
            overlay = hv.Overlay(results).collate()
            if len(markdown_results) == 0:
                return overlay
            return pn.Row(overlay, markdown_results)
        if len(markdown_results) > 0:
            return markdown_results
        return None

    def layout_plots(self, plot_callback: callable):
        if len(self.bench_cfg.result_vars) > 0:
            pt = hv.Layout()
            got_results = False
            for rv in self.bench_cfg.result_vars:
                res = plot_callback(rv)
                if res is not None:
                    got_results = True
                    pt += plot_callback(rv)
            return pt if got_results else None
        return plot_callback(self.bench_cfg.result_vars[0])

    def time_widget(self, title):
        return {"title": title}
        # if self.bench_cfg.over_time:
        #     time_widget_args = {"widget_type": "scrubber", "widget_location": "bottom"}
        #     time_widget_args["title"] = None  # use the title generated by the widget instead
        # else:
        #     time_widget_args = {"widget_type": "individual"}
        #     time_widget_args["title"] = title

        # return time_widget_args

    def to_bar(self, result_var: Parameter = None, **kwargs) -> Optional[pn.panel]:
        return self.filter(
            self.to_bar_ds,
            float_range=VarRange(0, 0),
            cat_range=VarRange(0, None),
            repeats_range=VarRange(1, 1),
            reduce=ReduceType.SQUEEZE,
            target_dimension=2,
            result_var=result_var,
            result_types=(ResultVar),
            **kwargs,
        )

    def to_bar_ds(self, dataset: xr.Dataset, result_var: Parameter = None, **kwargs):
        by = None
        if self.plt_cnt_cfg.cat_cnt >= 2:
            by = self.plt_cnt_cfg.cat_vars[1].name
        da_plot = dataset[result_var.name]
        title = self.title_from_ds(da_plot, result_var, **kwargs)
        time_widget_args = self.time_widget(title)
        return da_plot.hvplot.bar(by=by, **time_widget_args, **kwargs)

    def hv_container_ds(
        self, dataset: xr.Dataset, result_var: Parameter, container: hv.Chart = None, **kwargs
    ):
        return hv.Dataset(dataset[result_var.name]).to(container).opts(**kwargs)

    def to_hv_container(
        self,
        container: pn.pane.panel,
        reduce_type=ReduceType.AUTO,
        target_dimension: int = 2,
        result_var: Parameter = None,
        result_types=(ResultVar),
        **kwargs,
    ) -> Optional[pn.pane.panel]:
        return self.map_plot_panes(
            partial(self.hv_container_ds, container=container),
            hv_dataset=self.to_hv_dataset(reduce_type),
            target_dimension=target_dimension,
            result_var=result_var,
            result_types=result_types,
            **kwargs,
        )

    def to_line(self, result_var: Parameter = None, **kwargs) -> Optional[pn.panel]:
        return self.filter(
            self.to_line_ds,
            float_range=VarRange(1, 1),
            cat_range=VarRange(0, None),
            repeats_range=VarRange(1, 1),
            panel_range=VarRange(0, None),
            reduce=ReduceType.SQUEEZE,
            target_dimension=2,
            result_var=result_var,
            result_types=(ResultVar),
            **kwargs,
        )

    def to_line_ds(self, dataset: xr.Dataset, result_var: Parameter, **kwargs):
        x = self.plt_cnt_cfg.float_vars[0].name
        # y = self.plt_cnt_cfg.result_vars[0].name
        by = None
        if self.plt_cnt_cfg.cat_cnt >= 1:
            by = self.plt_cnt_cfg.cat_vars[0].name
        da_plot = dataset[result_var.name]
        title = self.title_from_ds(da_plot, result_var, **kwargs)
        time_widget_args = self.time_widget(title)
        return da_plot.hvplot.line(x=x, by=by, **time_widget_args, **kwargs)

    def to_curve(self, result_var: Parameter = None, **kwargs):
        return self.filter(
            self.to_curve_ds,
            float_range=VarRange(1, 1),
            cat_range=VarRange(0, None),
            repeats_range=VarRange(2, None),
            reduce=ReduceType.REDUCE,
            target_dimension=2,
            result_var=result_var,
            result_types=(ResultVar),
            **kwargs,
        )

    def to_curve_ds(
        self, dataset: xr.Dataset, result_var: Parameter, **kwargs
    ) -> Optional[hv.Curve]:
        hvds = hv.Dataset(dataset)
        # result_var = self.get_results_var_list(result_var)[0]
        title = self.title_from_ds(dataset, result_var, **kwargs)
        pt = hvds.to(hv.Curve).opts(title=title, **kwargs)
        pt *= hvds.to(hv.Spread).opts(alpha=0.2)
        if len(dataset.sizes) > 1:
            return pt.opts(legend_position="right").overlay()
        return pt.opts(legend_position="right")

    def to_heatmap(
        self,
        result_var: Parameter = None,
        tap_var=None,
        tap_container=None,
        target_dimension=2,
        **kwargs,
    ) -> Optional[pn.panel]:
        if tap_var is None:
            heatmap_cb = self.to_heatmap_ds
        else:
            heatmap_cb = partial(
                self.to_heatmap_container_tap_ds, result_var_plot=tap_var, container=tap_container
            )

        return self.filter(
            heatmap_cb,
            float_range=VarRange(2, None),
            cat_range=VarRange(0, None),
            input_range=VarRange(1, None),
            target_dimension=target_dimension,
            result_var=result_var,
            result_types=(ResultVar),
            **kwargs,
        )

    def to_heatmap_ds(
        self, dataset: xr.Dataset, result_var: Parameter, **kwargs
    ) -> Optional[hv.HeatMap]:
        if len(dataset.dims) >= 2:
            x = self.plt_cnt_cfg.float_vars[0].name
            y = self.plt_cnt_cfg.float_vars[1].name
            C = result_var.name
            title = f"Heatmap of {result_var.name}"
            time_args = self.time_widget(title)
            return dataset.hvplot.heatmap(x=x, y=y, C=C, cmap="plasma", **time_args, **kwargs)
        return None

    def to_heatmap_container_tap_ds(
        self,
        dataset: xr.Dataset,
        result_var: Parameter,
        result_var_plot: Parameter,
        container: pn.pane.panel = pn.pane.panel,
        **kwargs,
    ) -> pn.Row:
        htmap = self.to_heatmap_ds(dataset, result_var).opts(tools=["hover", "tap"], **kwargs)
        htmap_posxy = hv.streams.Tap(source=htmap, x=0, y=0)

        container_instance = container(**kwargs)
        title = pn.pane.Markdown("Selected: None")

        def tap_plot(x, y):  # pragma: no cover
            x_nearest = get_nearest_coords1D(
                x, dataset.coords[self.bench_cfg.input_vars[0].name].data
            )
            y_nearest = get_nearest_coords1D(
                y, dataset.coords[self.bench_cfg.input_vars[1].name].data
            )
            kdims = {}
            kdims[self.bench_cfg.input_vars[0].name] = x_nearest
            kdims[self.bench_cfg.input_vars[1].name] = y_nearest

            if hasattr(htmap, "current_key"):
                for d, k in zip(htmap.kdims, htmap.current_key):
                    kdims[d.name] = k

            ds = dataset[result_var_plot.name]
            val = ds.sel(**kdims)
            item = self.zero_dim_da_to_val(val)
            title.object = "Selected: " + ", ".join([f"{k}:{v}" for k, v in kdims.items()])
            container_instance.object = item
            if hasattr(container, "autoplay"):  # container is a video, set to autoplay
                container_instance.paused = False
                container_instance.time = 0
                container_instance.loop = True
                container_instance.autoplay = True

        htmap_posxy.add_subscriber(tap_plot)
        bound_plot = pn.Column(title, container_instance)
        return pn.Row(htmap, bound_plot)

    def to_error_bar(self) -> hv.Bars:
        return self.to_hv_dataset(ReduceType.REDUCE).to(hv.ErrorBars)

    def to_points(self, reduce: ReduceType = ReduceType.AUTO) -> hv.Points:
        ds = self.to_hv_dataset(reduce)
        pt = ds.to(hv.Points)
        if reduce:
            pt *= ds.to(hv.ErrorBars)
        return pt

    def to_scatter(self, **kwargs) -> Optional[pn.panel]:
        match_res = PlotFilter(
            float_range=VarRange(0, 0), cat_range=VarRange(0, None), repeats_range=VarRange(1, 1)
        ).matches_result(self.plt_cnt_cfg, "to_hvplot_scatter")
        if match_res.overall:
            hv_ds = self.to_hv_dataset(ReduceType.SQUEEZE)
            by = None
            subplots = False
            if self.plt_cnt_cfg.cat_cnt > 1:
                by = [v.name for v in self.bench_cfg.input_vars[1:]]
                subplots = False
            return hv_ds.data.hvplot.scatter(by=by, subplots=subplots, **kwargs).opts(
                title=self.to_plot_title()
            )
        return match_res.to_panel(**kwargs)

    # def to_scatter_jitter(self, **kwargs) -> Optional[hv.Scatter]:
    #     matches = PlotFilter(
    #         float_range=VarRange(0, 0),
    #         cat_range=VarRange(0, None),
    #         repeats_range=VarRange(2, None),
    #         input_range=VarRange(1, None),
    #     ).matches_result(self.plt_cnt_cfg, "to_scatter_jitter")
    #     if matches.overall:
    #         hv_ds = self.to_hv_dataset(ReduceType.NONE)

    #         by = None
    #         groupby = None
    #         subplots=False
    #         if self.plt_cnt_cfg.cat_cnt > 1:
    #             by = [v.name for v in self.bench_cfg.all_vars[1:]]
    #             subplots=False
    #         return hv_ds.data.hvplot.scatter(by=by,subplots=subplots, **kwargs).opts(title=self.to_plot_title())

    #         # pt = (
    #         #     hv_ds.to(hv.Scatter)
    #         #     .opts(jitter=0.1)
    #         #     .overlay("repeat")
    #         #     .opts(show_legend=False, title=self.to_plot_title(), **kwargs)
    #         # )
    #         # return pt
    #     return matches.to_panel()

    # def to_scatter_jitter_multi(self, **kwargs) -> Optional[hv.Scatter]:
    #     matches = PlotFilter(
    #         float_range=VarRange(0, 0),
    #         cat_range=VarRange(0, None),
    #         repeats_range=VarRange(2, None),
    #         input_range=VarRange(1, None),
    #     ).matches_result(self.plt_cnt_cfg, "to_scatter_jitter")
    #     if matches.overall:
    #         hv_dataset = self.to_hv_dataset(ReduceType.NONE)

    #         print("KDIMS",hv_dataset.kdims)
    #         # hv_dataset.kdims =[hv_dataset.kdims[2],hv_dataset.kdims[1],hv_dataset.kdims[0]]
    #         # print("KDIMS",hv_dataset.kdims)

    #         # exit()
    #         cb = partial(self.to_scatter_jitter_da, **kwargs)
    #         return self.to_panes_multi_panel(hv_dataset, None, plot_callback=cb, target_dimension=3)
    #     return matches.to_panel()

    # def to_scatter_jitter_da(self, ds: xr.Dataset, **kwargs) -> Optional[hv.Scatter]:
    #     matches = PlotFilter(
    #         float_range=VarRange(0, 0),
    #         cat_range=VarRange(0, None),
    #         repeats_range=VarRange(2, None),
    #         input_range=VarRange(1, None),
    #     ).matches_result(self.plt_cnt_cfg, "to_scatter_jitter")
    #     if matches.overall:

    #         print("DA IN",da)
    #         da = self.to_hv_dataset(ReduceType.NONE)
    #         hvds = hv.Dataset(da)
    #         # return None

    #         # print("DA FRESH",da)
    #         result_var = self.bench_cfg.result_vars[0]

    #         print(hvds.data.sizes)
    #         print("repeat" in hvds.data.sizes)
    #         # if "repeat" in hvds.data.sizes:
    #         # try:
    #         #     pt = (
    #         #         hvds.to(hv.Scatter)
    #         #         .opts(jitter=0.1)
    #         #         # .overlay()
    #         #         .overlay("repeat")
    #         #         .opts(show_legend=False, title=self.to_plot_title(), clabel=result_var.name, **kwargs)
    #         #     )
    #         # except:
    #         pt = (
    #             hvds.to(hv.Scatter)
    #             .opts(jitter=0.1)
    #             # .overlay()
    #             # .overlay("repeat")
    #             .opts(show_legend=False, title=self.to_plot_title(), clabel=result_var.name, **kwargs)
    #         )
    #         return pt
    #     return matches.to_panel()

    def to_scatter_jitter(
        self,
        result_var: Parameter = None,
        **kwargs,  # pylint: disable=unused-argument
    ) -> List[hv.Scatter]:
        return self.overlay_plots(partial(self.to_scatter_jitter_single, **kwargs))

    def to_scatter_jitter_single(self, result_var: Parameter, **kwargs) -> Optional[hv.Scatter]:
        matches = PlotFilter(
            float_range=VarRange(0, 0),
            cat_range=VarRange(0, None),
            repeats_range=VarRange(2, None),
            input_range=VarRange(1, None),
        ).matches_result(self.plt_cnt_cfg, "to_scatter_jitter")
        if matches.overall:
            ds = self.to_hv_dataset(ReduceType.NONE)
            pt = (
                ds.to(hv.Scatter, vdims=[result_var.name], label=result_var.name)
                .opts(jitter=0.1, show_legend=False, title=self.to_plot_title(), **kwargs)
                .overlay("repeat")
            )
            return pt
        return matches.to_panel()

    def to_heatmap_single(
        self, result_var: Parameter, reduce: ReduceType = ReduceType.AUTO, **kwargs
    ) -> hv.HeatMap:
        matches_res = PlotFilter(
            float_range=VarRange(2, None),
            cat_range=VarRange(0, None),
            input_range=VarRange(1, None),
        ).matches_result(self.plt_cnt_cfg, "to_heatmap")
        if matches_res.overall:
            z = result_var
            title = f"{z.name} vs ("

            for iv in self.bench_cfg.input_vars:
                title += f" vs {iv.name}"
            title += ")"

            color_label = f"{z.name} [{z.units}]"

            return self.to(hv.HeatMap, reduce).opts(clabel=color_label, **kwargs)
        return matches_res.to_panel()

    def to_heatmap_tap(
        self,
        result_var: Parameter,
        reduce: ReduceType = ReduceType.AUTO,
        width=800,
        height=800,
        **kwargs,
    ):
        htmap = self.to_heatmap_single(result_var, reduce).opts(
            tools=["hover", "tap"], width=width, height=height
        )
        htmap_posxy = hv.streams.Tap(source=htmap, x=0, y=0)

        def tap_plot(x, y):
            kwargs[self.bench_cfg.input_vars[0].name] = x
            kwargs[self.bench_cfg.input_vars[1].name] = y
            return self.get_nearest_holomap(**kwargs).opts(width=width, height=height)

        tap_htmap = hv.DynamicMap(tap_plot, streams=[htmap_posxy])
        return htmap + tap_htmap

    def to_nd_layout(self, hmap_name: str) -> hv.NdLayout:
        return hv.NdLayout(self.get_hmap(hmap_name), kdims=self.bench_cfg.hmap_kdims).opts(
            shared_axes=False, shared_datasource=False
        )

    def to_holomap(self, name: str = None) -> hv.HoloMap:
        return hv.HoloMap(self.to_nd_layout(name)).opts(shared_axes=False)

    def to_holomap_list(self, hmap_names: List[str] = None) -> hv.HoloMap:
        if hmap_names is None:
            hmap_names = [i.name for i in self.result_hmaps]
        col = pn.Column()
        for name in hmap_names:
            self.to_holomap(name)
        return col

    def get_nearest_holomap(self, name: str = None, **kwargs):
        canonical_inp = hmap_canonical_input(
            get_nearest_coords(self.ds, collapse_list=True, **kwargs)
        )
        return self.get_hmap(name)[canonical_inp].opts(framewise=True)

    def to_dynamic_map(self, name: str = None) -> hv.DynamicMap:
        """use the values stored in the holomap dictionary to populate a dynamic map. Note that this is much faster than passing the holomap to a holomap object as the values are calculated on the fly"""

        def cb(**kwargs):
            return self.get_hmap(name)[hmap_canonical_input(kwargs)].opts(
                framewise=True, shared_axes=False
            )

        kdims = []
        for i in self.bench_cfg.input_vars + [self.bench_cfg.iv_repeat]:
            kdims.append(i.as_dim(compute_values=True))

        return hv.DynamicMap(cb, kdims=kdims)

    def to_grid(self, inputs=None):
        if inputs is None:
            inputs = self.bench_cfg.inputs_as_str()
        if len(inputs) > 2:
            inputs = inputs[:2]
        return self.to_holomap().grid(inputs)

    def to_table(self):
        return self.to(hv.Table, ReduceType.SQUEEZE)

    def to_surface(self, result_var: Parameter = None, **kwargs) -> Optional[pn.panel]:
        return self.filter(
            self.to_surface_ds,
            float_range=VarRange(2, None),
            cat_range=VarRange(0, None),
            input_range=VarRange(1, None),
            reduce=ReduceType.REDUCE,
            target_dimension=2,
            result_var=result_var,
            result_types=(ResultVar),
            **kwargs,
        )

    def to_surface_ds(
        self, dataset: xr.Dataset, result_var: Parameter, alpha: float = 0.3, **kwargs
    ) -> Optional[pn.panel]:
        """Given a benchCfg generate a 2D surface plot

        Args:
            result_var (Parameter): result variable to plot

        Returns:
            pn.pane.holoview: A 2d surface plot as a holoview in a pane
        """
        matches_res = PlotFilter(
            float_range=VarRange(2, 2),
            cat_range=VarRange(0, None),
            vector_len=VarRange(1, 1),
            result_vars=VarRange(1, 1),
        ).matches_result(self.plt_cnt_cfg, "to_surface_hv")
        if matches_res.overall:
            # xr_cfg = plot_float_cnt_2(self.plt_cnt_cfg, result_var)

            # TODO a warning suggests setting this parameter, but it does not seem to help as expected, leaving here to fix in the future
            # hv.config.image_rtol = 1.0

            mean = dataset[result_var.name]

            hvds = hv.Dataset(dataset[result_var.name])

            x = self.plt_cnt_cfg.float_vars[0]
            y = self.plt_cnt_cfg.float_vars[1]

            try:
                surface = hvds.to(hv.Surface, vdims=[result_var.name])
                surface = surface.opts(colorbar=True)
            except Exception as e:
                logging.warning(e)

            if self.bench_cfg.repeats > 1:
                std_dev = dataset[f"{result_var.name}_std"]

                upper = mean + std_dev
                upper.name = result_var.name

                lower = mean - std_dev
                lower.name = result_var.name

                surface *= (
                    hv.Dataset(upper)
                    .to(hv.Surface)
                    .opts(alpha=alpha, colorbar=False, backend="plotly")
                )
                surface *= (
                    hv.Dataset(lower)
                    .to(hv.Surface)
                    .opts(alpha=alpha, colorbar=False, backend="plotly")
                )

            surface = surface.opts(
                zlabel=f"{result_var.name} [{result_var.units}]",
                title=f"{result_var.name} vs ({x.name} and {y.name})",
                backend="plotly",
                **kwargs,
            )

            if self.bench_cfg.render_plotly:
                hv.extension("plotly")
                out = surface
            else:
                # using render disabled the holoviews sliders :(
                out = hv.render(surface, backend="plotly")
            return pn.Column(out, name="surface_hv")

        return matches_res.to_panel()

    # def plot_scatter2D_hv(self, rv: ParametrizedSweep) -> pn.pane.Plotly:
    # import plotly.express as px

    #     """Given a benchCfg generate a 2D scatter plot

    #     Args:
    #         bench_cfg (BenchCfg): description of benchmark
    #         rv (ParametrizedSweep): result variable to plot

    #     Returns:
    #         pn.pane.Plotly: A 3d volume plot as a holoview in a pane
    #     """

    #     # bench_cfg = wrap_long_time_labels(bench_cfg)
    #     self.ds.drop_vars("repeat")

    #     df = self.to_pandas()

    #     names = rv.index_names()

    #     return px.scatter(
    #         df, x=names[0], y=names[1], marginal_x="histogram", marginal_y="histogram"
    #     )


HoloviewResult.set_default_opts()



import logging
from typing import Callable
import os
import panel as pn
from pathlib import Path
import tempfile

from threading import Thread

from bencher.results.bench_result import BenchResult
from bencher.bench_plot_server import BenchPlotServer
from bencher.bench_cfg import BenchRunCfg


class BenchReport(BenchPlotServer):
    def __init__(
        self,
        bench_name: str = None,
    ) -> None:
        self.bench_name = bench_name
        self.pane = pn.Tabs(tabs_location="left", name=self.bench_name)

    def append_title(self, title: str, new_tab: bool = True):
        if new_tab:
            return self.append_tab(pn.pane.Markdown(f"# {title}", name=title), title)
        return self.append_markdown(f"# {title}", title)

    def append_markdown(self, markdown: str, name=None, width=800, **kwargs) -> pn.pane.Markdown:
        if name is None:
            name = markdown
        md = pn.pane.Markdown(markdown, name=name, width=width, **kwargs)
        self.append(md, name)
        return md

    def append(self, pane: pn.panel, name: str = None) -> None:
        if len(self.pane) == 0:
            if name is None:
                name = pane.name
            self.append_tab(pane, name)
        else:
            self.pane[-1].append(pane)

    def append_col(self, pane: pn.panel, name: str = None) -> None:
        if name is not None:
            col = pn.Column(pane, name=name)
        else:
            col = pn.Column(pane, name=pane.name)
        self.pane.append(col)

    def append_result(self, bench_res: BenchResult) -> None:
        self.append_tab(bench_res.plot(), bench_res.bench_cfg.title)

    def append_tab(self, pane: pn.panel, name: str = None) -> None:
        if pane is not None:
            if name is None:
                name = pane.name
            self.pane.append(pn.Column(pane, name=name))

    def save_index(self, directory="", filename="index.html") -> Path:
        """Saves the result to index.html in the root folder so that it can be displayed by github pages.

        Returns:
            Path: save path
        """
        return self.save(directory, filename, False)

    def save(
        self,
        directory: str | Path = "cachedir",
        filename: str = None,
        in_html_folder: bool = True,
        **kwargs,
    ) -> Path:
        """Save the result to a html file.  Note that dynamic content will not work.  by passing save(__file__) the html output will be saved in the same folder as the source code in a html subfolder.

        Args:
            directory (str | Path, optional): base folder to save to. Defaults to "cachedir" which should be ignored by git.
            filename (str, optional): The name of the html file. Defaults to the name of the benchmark
            in_html_folder (bool, optional): Put the saved files in a html subfolder to help keep the results separate from source code. Defaults to True.

        Returns:
            Path: the save path
        """

        if filename is None:
            filename = f"{self.bench_name}.html"

        base_path = Path(directory)

        if in_html_folder:
            base_path /= "html"

        logging.info(f"creating dir {base_path.absolute()}")
        os.makedirs(base_path.absolute(), exist_ok=True)

        base_path = base_path / filename

        logging.info(f"saving html output to: {base_path.absolute()}")

        self.pane.save(filename=base_path, progress=True, embed=True, **kwargs)
        return base_path

    def show(self, run_cfg: BenchRunCfg = None) -> Thread:  # pragma: no cover
        """Launches a webserver with plots of the benchmark results, blocking

        Args:
            run_cfg (BenchRunCfg, optional): Options for the webserve such as the port. Defaults to None.

        """
        if run_cfg is None:
            run_cfg = BenchRunCfg()

        return BenchPlotServer().plot_server(self.bench_name, run_cfg, self.pane)

    def publish(
        self, remote_callback: Callable, branch_name: str = None, debug: bool = False
    ) -> str:  # pragma: no cover
        """Publish the results as an html file by committing it to the bench_results branch in the current repo. If you have set up your repo with github pages or equivalent then the html file will be served as a viewable webpage.  This is an example of a callable to publish on github pages:

        .. code-block:: python

            def publish_args(branch_name) -> Tuple[str, str]:
                return (
                    "https://github.com/dyson-ai/bencher.git",
                    f"https://github.com/dyson-ai/bencher/blob/{branch_name}")


        Args:
            remote (Callable): A function the returns a tuple of the publishing urls. It must follow the signature def publish_args(branch_name) -> Tuple[str, str].  The first url is the git repo name, the second url needs to match the format for viewable html pages on your git provider.  The second url can use the argument branch_name to point to the report on a specified branch.

        Returns:
            str: the url of the published report
        """

        if branch_name is None:
            branch_name = self.bench_name
        branch_name += "_debug" if debug else ""

        remote, publish_url = remote_callback(branch_name)

        with tempfile.TemporaryDirectory() as td:
            directory = td
            report_path = self.save(directory, filename="index.html", in_html_folder=False)
            logging.info(f"created report at: {report_path.absolute()}")

            cd_dir = f"cd {directory} &&"

            os.system(f"{cd_dir} git init")
            os.system(f"{cd_dir} git checkout -b {branch_name}")
            os.system(f"{cd_dir} git add index.html")
            os.system(f'{cd_dir} git commit -m "publish {branch_name}"')
            os.system(f"{cd_dir} git remote add origin {remote}")
            os.system(f"{cd_dir} git push --set-upstream origin {branch_name} -f")

        logging.info("Published report @")
        logging.info(publish_url)

        return publish_url


# def append(self,pane):
# self.report.append(pane)

# def __getstate__(self):
#     state = self.__dict__.copy()
#     # Don't pickle baz
#     del state["pane"]
#     return state

# def __setstate__(self, state):
#     self.__dict__.update(state)
#     # Add baz back since it doesn't exist in the pickle
#     self.report = []

# def publish_old(
#     self,
#     directory: str = "bench_results",
#     branch_name: str = "bench_results",
#     url_postprocess: Callable = None,
#     **kwargs,
# ) -> str:
#     """Publish the results as an html file by committing it to the bench_results branch in the current repo. If you have set up your repo with github pages or equivalent then the html file will be served as a viewable webpage.

#     Args:
#         directory (str, optional): Directory to save the results. Defaults to "bench_results".
#         branch_name (str, optional): Branch to publish on. Defaults to "bench_results".
#         url_postprocess (Callable, optional): A function that maps the origin url to a github pages url. Pass your own function if you are using another git providers. Defaults to None.

#     Returns:
#         str: _description_
#     """

#     def get_output(cmd: str) -> str:
#         return (
#             subprocess.run(cmd.split(" "), stdout=subprocess.PIPE, check=False)
#             .stdout.decode("utf=8")
#             .strip()
#         )

#     def postprocess_url(publish_url: str, branch_name: str, report_path: str, **kwargs) -> str:
#         # import re

#         # return re.sub(
#         #     """((git|ssh|http(s)?)|(git@[\w\.-]+))(:(//)?)([\w\.@\:/\-~]+)(\.git)(/)?""",
#         #     """https://$7/""",
#         #     publish_url,
#         # )
#         # git@github.com:user/project.git
#         # https://github.com/user/project.git
#         # http://github.com/user/project.git
#         # git@192.168.101.127:user/project.git
#         # https://192.168.101.127/user/project.git
#         # http://192.168.101.127/user/project.git
#         # ssh://user@host.xz:port/path/to/repo.git/
#         # ssh://user@host.xz/path/to/repo.git/
#         # ssh://host.xz:port/path/to/repo.git/
#         # ssh://host.xz/path/to/repo.git/
#         # ssh://user@host.xz/path/to/repo.git/
#         # ssh://host.xz/path/to/repo.git/
#         # ssh://user@host.xz/~user/path/to/repo.git/
#         # ssh://host.xz/~user/path/to/repo.git/
#         # ssh://user@host.xz/~/path/to/repo.git
#         # ssh://host.xz/~/path/to/repo.git
#         # git://host.xz/path/to/repo.git/
#         # git://host.xz/~user/path/to/repo.git/
#         # http://host.xz/path/to/repo.git/
#         # https://host.xz/path/to/repo.git/
#         # https://regex101.com/r/qT7NP0/3

#         return publish_url.replace(".git", f"/blob/{directory}/{report_path}")

#     if url_postprocess is None:
#         url_postprocess = postprocess_url
#     current_branch = get_output("git symbolic-ref --short HEAD")
#     logging.info(f"on branch: {current_branch}")
#     stash_msg = get_output("git stash")
#     logging.info(f"stashing current work :{stash_msg}")
#     checkout_msg = get_output(f"git checkout -b {branch_name}")
#     checkout_msg = get_output(f"git checkout {branch_name}")
#     get_output("git pull")

#     logging.info(f"checking out branch: {checkout_msg}")
#     report_path = self.save(directory, in_html_folder=False)
#     logging.info(f"created report at: {report_path.absolute()}")
#     # commit_msg = f""
#     logging.info("adding report to git")
#     get_output(f"git add {report_path.absolute()}")
#     get_output("git status")
#     logging.info("committing report")
#     cmd = f'git commit -m "generate_report:{self.bench_name}"'
#     logging.info(cmd)
#     get_output(cmd)
#     logging.info("pushing report to origin")
#     get_output(f"git push --set-upstream origin {branch_name}")
#     logging.info("checking out original branch")
#     get_output(f"git checkout {current_branch}")
#     if "No local changes" not in stash_msg:
#         logging.info("restoring work with git stash pop")
#         get_output("git stash pop")

#     publish_url = get_output("git remote get-url --push origin")
#     logging.info(f"raw url:{publish_url}")
#     publish_url = url_postprocess(
#         publish_url, branch_name=branch_name, report_path=report_path, **kwargs
#     )
#     logging.info("Published report @")
#     logging.info(publish_url)
#     return publish_url
